import datetime
import os
import sys

import pytest
from dagster._core.definitions.run_request import InstigatorType
from dagster._core.definitions.sensor_definition import SensorType
from dagster._core.remote_origin import InProcessCodeLocationOrigin, RemoteRepositoryOrigin
from dagster._core.remote_representation.external import CompoundID
from dagster._core.scheduler.instigation import (
    InstigatorState,
    InstigatorStatus,
    SensorInstigatorData,
    TickData,
    TickStatus,
)
from dagster._core.storage.dagster_run import DagsterRun
from dagster._core.storage.tags import RUN_KEY_TAG, SENSOR_NAME_TAG
from dagster._core.test_utils import SingleThreadPoolExecutor, freeze_time, wait_for_futures
from dagster._core.types.loadable_target_origin import LoadableTargetOrigin
from dagster._core.utils import make_new_backfill_id, make_new_run_id
from dagster._core.workspace.context import WorkspaceRequestContext
from dagster._daemon import get_default_daemon_logger
from dagster._daemon.sensor import execute_sensor_iteration
from dagster._time import get_timezone
from dagster._utils import Counter, traced_counter
from dagster._utils.error import SerializableErrorInfo
from dagster._vendored.dateutil.relativedelta import relativedelta
from dagster_graphql.implementation.utils import UserFacingGraphQLError
from dagster_graphql.schema.instigation import GrapheneDynamicPartitionsRequestType
from dagster_graphql.test.utils import (
    execute_dagster_graphql,
    infer_repository_selector,
    infer_sensor_selector,
    main_repo_location_name,
    main_repo_name,
)

from dagster_graphql_tests.graphql.graphql_context_test_suite import (
    ExecutingGraphQLContextTestMatrix,
    NonLaunchableGraphQLContextTestMatrix,
    ReadonlyGraphQLContextTestMatrix,
)

GET_SENSORS_QUERY = """
query SensorsQuery($repositorySelector: RepositorySelector!) {
  sensorsOrError(repositorySelector: $repositorySelector) {
    __typename
    ... on PythonError {
      message
      stack
    }
    ... on Sensors {
      results {
        name
        targets {
          pipelineName
          solidSelection
          mode
        }
        description
        minIntervalSeconds
        sensorState {
          status
          runs {
              id
              runId
          }
          runsCount
          ticks {
              id
              status
              timestamp
              runIds
              error {
                  message
                  stack
              }
              skipReason
          }
        }
      }
    }
  }
}
"""

GET_SENSORS_BY_STATUS_QUERY = """
query SensorsByStatusQuery($repositorySelector: RepositorySelector!, $status: InstigationStatus) {
  sensorsOrError(repositorySelector: $repositorySelector, sensorStatus: $status) {
    __typename
    ... on PythonError {
      message
      stack
    }
    ... on Sensors {
      results {
        name
        sensorState {
          status
        }
      }
    }
  }
}
"""

GET_SENSORS_TICKS_BY_STATUS_QUERY = """
query SensorsTicksByStatusQuery($repositorySelector: RepositorySelector!, $statuses: [InstigationTickStatus!]) {
  sensorsOrError(repositorySelector: $repositorySelector) {
    __typename
    ... on PythonError {
      message
      stack
    }
    ... on Sensors {
      results {
        name
        sensorState {
          ticks(statuses: $statuses) {
            status
          }
        }
      }
    }
  }
}
"""


GET_SENSOR_QUERY = """
query SensorQuery($sensorSelector: SensorSelector!) {
  sensorOrError(sensorSelector: $sensorSelector) {
    __typename
    ... on PythonError {
      message
      stack
    }
    ... on Sensor {
      name
      targets {
        pipelineName
        solidSelection
        mode
      }
      minIntervalSeconds
      nextTick {
        timestamp
      }
      owners {
        ... on UserDefinitionOwner {
          email
        }
        ... on TeamDefinitionOwner {
          team
        }
      }
      sensorState {
        status
        runs {
          id
          runId
        }
        runsCount
        ticks {
            id
            status
            timestamp
            runIds
            error {
                message
                stack
                errorChain {
                    error {
                        message
                        stack
                    }
                }
            }
        }
      }
      sensorType
      tags {
        key
        value
      }
      metadataEntries {
        label
        ... on TextMetadataEntry {
          text
        }
      }
      assetSelection {
        assetSelectionString
        assetKeys {
          path
        }
        assets {
          key {
            path
          }
          definition {
            assetKey {
              path
            }
          }
        }
        assetsOrError {
          ... on AssetConnection {
            nodes {
              key {
                path
              }
            }
          }
          ... on PythonError {
            message
          }
        }
      }
    }
  }
}
"""


GET_ASSET_SELECTION_ERROR_QUERY = """
query SensorQuery($sensorSelector: SensorSelector!) {
  sensorOrError(sensorSelector: $sensorSelector) {
    __typename
    ... on PythonError {
      message
      stack
    }
    ... on Sensor {
      name
      assetSelection {
        assetSelectionString
        assetsOrError {
          __typename
          ... on AssetConnection {
            nodes {
              key {
                path
              }
            }
          }
          ... on PythonError {
            message
          }
        }
      }
    }
  }
}
"""


GET_SENSOR_STATUS_QUERY = """
query SensorStateQuery($sensorSelector: SensorSelector!) {
  sensorOrError(sensorSelector: $sensorSelector) {
    __typename
    ... on Sensor {
      canReset
      defaultStatus
      hasCursorUpdatePermissions
      sensorState {
        id
        status
        selectorId
        hasStartPermission
        hasStopPermission
      }
    }
  }
}
"""

GET_SENSOR_TICK_RANGE_QUERY = """
query SensorQuery($sensorSelector: SensorSelector!, $dayRange: Int, $dayOffset: Int, $beforeTimestamp: Float, $afterTimestamp: Float) {
  sensorOrError(sensorSelector: $sensorSelector) {
    __typename
    ... on PythonError {
      message
      stack
    }
    ... on Sensor {
      id
      sensorState {
        id
        ticks(dayRange: $dayRange, dayOffset: $dayOffset, beforeTimestamp: $beforeTimestamp, afterTimestamp: $afterTimestamp) {
          id
          timestamp
          endTimestamp
        }
      }
    }
  }
}
"""

START_SENSORS_QUERY = """
mutation($sensorSelector: SensorSelector!) {
  startSensor(sensorSelector: $sensorSelector) {
    __typename
    ... on PythonError {
      message
      className
      stack
    }
    ... on Sensor {
      id
      canReset
      sensorState {
        selectorId
        status
      }
    }
  }
}
"""

STOP_SENSORS_QUERY = """
mutation(
  $id: String
  $jobOriginId: String
  $jobSelectorId: String
) {
  stopSensor(
    id: $id
    jobOriginId: $jobOriginId
    jobSelectorId: $jobSelectorId
  ) {
    __typename
    ... on PythonError {
      message
      className
      stack
    }
    ... on StopSensorMutationResult {
      instigationState {
        status
      }
    }
  }
}
"""

RESET_SENSORS_QUERY = """
mutation($sensorSelector: SensorSelector!) {
  resetSensor(sensorSelector: $sensorSelector) {
    __typename
    ... on PythonError {
      message
      className
      stack
    }
    ... on Sensor {
      id
      canReset
      sensorState {
        selectorId
        status
      }
    }
  }
}
"""

GET_SENSOR_CURSOR_QUERY = """
query SensorCursorQuery($sensorSelector: SensorSelector!) {
  sensorOrError(sensorSelector: $sensorSelector) {
    __typename
    ... on Sensor {
      sensorState {
        id
        status
        typeSpecificData {
          ... on SensorData {
            lastCursor
          }
        }
      }
    }
  }
}
"""

SET_SENSOR_CURSOR_MUTATION = """
mutation($sensorSelector: SensorSelector!, $cursor: String) {
  setSensorCursor(sensorSelector: $sensorSelector, cursor: $cursor) {
    __typename
    ... on PythonError {
      message
      className
      stack
    }
    ... on Sensor {
      id
      sensorState {
        status
        typeSpecificData {
          ... on SensorData {
            lastCursor
          }
        }
      }
    }
  }
}
"""

SENSOR_DRY_RUN_MUTATION = """
mutation($selectorData: SensorSelector!, $cursor: String) {
  sensorDryRun(selectorData: $selectorData, cursor: $cursor) {
    __typename
    ... on PythonError {
      message
      stack
    }
    ... on DryRunInstigationTick {
      evaluationResult {
        cursor
        runRequests {
          runConfigYaml
        }
        skipReason
        error {
          message
          stack
        }
        dynamicPartitionsRequests {
          partitionKeys
          partitionsDefName
          type
        }
      }
    }
    ... on SensorNotFoundError {
      sensorName
    }
  }
}
"""

REPOSITORY_SENSORS_QUERY = """
query RepositorySensorsQuery($repositorySelector: RepositorySelector!, $sensorType: SensorType) {
    repositoryOrError(repositorySelector: $repositorySelector) {
        ... on Repository {
            id
            sensors(sensorType: $sensorType) {
                id
                name
                sensorType
                sensorState {
                    id
                    runs(limit: 1) {
                      id
                      runId
                    }
                    ticks(limit: 1) {
                      id
                      timestamp
                    }
                }
            }
        }
    }
}
"""


GET_TICKS_QUERY = """
query TicksQuery($sensorSelector: SensorSelector!, $statuses: [InstigationTickStatus!], $tickId: ID!) {
  sensorOrError(sensorSelector: $sensorSelector) {
    __typename
    ... on PythonError {
      message
      stack
    }
    ... on Sensor {
      id
      sensorState {
        id
        tick(tickId: $tickId) {
          tickId
        }
        ticks(statuses: $statuses) {
          id
          tickId
          status
          timestamp
          runs {
            id
          }
        }
      }
    }
  }
}
"""

GET_TICK_LOGS_QUERY = """
query TickLogsQuery($sensorSelector: SensorSelector!) {
  sensorOrError(sensorSelector: $sensorSelector) {
    __typename
    ... on PythonError {
      message
      stack
    }
    ... on Sensor {
      id
      sensorState {
        id
        ticks {
          id
          logEvents {
            events {
              message
            }
          }
        }
      }
    }
  }
}
"""

GET_TICK_DYNAMIC_PARTITIONS_REQUEST_RESULTS_QUERY = """
query TickDynamicPartitionsRequestResultsQuery($sensorSelector: SensorSelector!) {
  sensorOrError(sensorSelector: $sensorSelector) {
    __typename
    ... on PythonError {
      message
      stack
    }
    ... on Sensor {
      id
      sensorState {
        id
        ticks {
          id
          dynamicPartitionsRequestResults {
            partitionsDefName
            partitionKeys
            skippedPartitionKeys
            type
          }
        }
      }
    }
  }
}
"""

INSTIGATION_STATE_QUERY = """
query InstigationStateQuery($instigationSelector: InstigationSelector! $id: String) {
  instigationStateOrError(instigationSelector: $instigationSelector id: $id) {
    __typename
    ... on PythonError {
      message
      stack
    }
    ... on InstigationState {
        id
        status
    }
  }
}
"""


class TestSensors(NonLaunchableGraphQLContextTestMatrix):
    @pytest.mark.parametrize(
        "sensor_name, expected_type",
        [
            ("always_no_config_sensor_with_tags_and_metadata", "STANDARD"),
            ("run_status", "RUN_STATUS"),
            ("single_asset_sensor", "ASSET"),
            ("many_asset_sensor", "MULTI_ASSET"),
            ("the_failure_sensor", "RUN_STATUS"),
        ],
    )
    def test_sensor_types(
        self, graphql_context: WorkspaceRequestContext, sensor_name, expected_type
    ):
        sensor_selector = infer_sensor_selector(graphql_context, sensor_name)
        result = execute_dagster_graphql(
            graphql_context,
            GET_SENSOR_QUERY,
            variables={"sensorSelector": sensor_selector},
        )

        assert result.data
        assert result.data["sensorOrError"]
        assert result.data["sensorOrError"]["__typename"] == "Sensor"
        sensor = result.data["sensorOrError"]
        assert sensor["sensorType"] == expected_type

    def test_dry_run(self, graphql_context: WorkspaceRequestContext):
        instigator_selector = infer_sensor_selector(
            graphql_context, "always_no_config_sensor_with_tags_and_metadata"
        )
        result = execute_dagster_graphql(
            graphql_context,
            SENSOR_DRY_RUN_MUTATION,
            variables={"selectorData": instigator_selector, "cursor": "blah"},
        )
        assert result.data
        assert result.data["sensorDryRun"]["__typename"] == "DryRunInstigationTick"
        evaluation_result = result.data["sensorDryRun"]["evaluationResult"]
        assert evaluation_result["cursor"] == "blah"
        assert len(evaluation_result["runRequests"]) == 1
        assert evaluation_result["runRequests"][0]["runConfigYaml"] == "{}\n"
        assert evaluation_result["skipReason"] is None
        assert evaluation_result["error"] is None
        assert evaluation_result["dynamicPartitionsRequests"] == []

    def test_dry_run_with_run_key(self, graphql_context: WorkspaceRequestContext):
        instigator_selector = infer_sensor_selector(graphql_context, "run_key_sensor")
        result = execute_dagster_graphql(
            graphql_context,
            SENSOR_DRY_RUN_MUTATION,
            variables={"selectorData": instigator_selector, "cursor": "blah"},
        )

        assert result.data
        assert result.data["sensorDryRun"]["__typename"] == "DryRunInstigationTick"
        evaluation_result = result.data["sensorDryRun"]["evaluationResult"]
        assert evaluation_result["cursor"] == "blah"
        assert len(evaluation_result["runRequests"]) == 1
        assert evaluation_result["runRequests"][0]["runConfigYaml"] == "{}\n"
        assert evaluation_result["skipReason"] is None
        assert evaluation_result["error"] is None
        assert evaluation_result["dynamicPartitionsRequests"] == []

        graphql_context.instance.add_run(
            DagsterRun(
                job_name="run_key_sensor",
                run_id=make_new_run_id(),
                tags={RUN_KEY_TAG: "the_key", SENSOR_NAME_TAG: "run_key_sensor"},
            )
        )

        result = execute_dagster_graphql(
            graphql_context,
            SENSOR_DRY_RUN_MUTATION,
            variables={"selectorData": instigator_selector, "cursor": "blah"},
        )
        assert result.data
        assert result.data["sensorDryRun"]["__typename"] == "DryRunInstigationTick"
        evaluation_result = result.data["sensorDryRun"]["evaluationResult"]
        # no more run run requests because the key matches
        assert len(evaluation_result["runRequests"]) == 0

    def test_dry_run_with_dynamic_partition_requests(
        self, graphql_context: WorkspaceRequestContext
    ):
        instigator_selector = infer_sensor_selector(
            graphql_context, "dynamic_partition_requesting_sensor"
        )
        result = execute_dagster_graphql(
            graphql_context,
            SENSOR_DRY_RUN_MUTATION,
            variables={"selectorData": instigator_selector, "cursor": "blah"},
        )
        assert result.data
        assert result.data["sensorDryRun"]["__typename"] == "DryRunInstigationTick"
        evaluation_result = result.data["sensorDryRun"]["evaluationResult"]
        assert evaluation_result["cursor"] == "blah"
        assert len(evaluation_result["runRequests"]) == 1
        assert evaluation_result["runRequests"][0]["runConfigYaml"] == "{}\n"
        assert evaluation_result["skipReason"] is None
        assert evaluation_result["error"] is None
        assert len(evaluation_result["dynamicPartitionsRequests"]) == 2
        assert evaluation_result["dynamicPartitionsRequests"][0]["partitionKeys"] == [
            "new_key",
            "new_key2",
            "existent_key",
        ]
        assert evaluation_result["dynamicPartitionsRequests"][0]["partitionsDefName"] == "foo"
        assert (
            evaluation_result["dynamicPartitionsRequests"][0]["type"]
            == GrapheneDynamicPartitionsRequestType.ADD_PARTITIONS
        )
        assert evaluation_result["dynamicPartitionsRequests"][1]["partitionKeys"] == [
            "old_key",
            "nonexistent_key",
        ]
        assert evaluation_result["dynamicPartitionsRequests"][1]["partitionsDefName"] == "foo"
        assert (
            evaluation_result["dynamicPartitionsRequests"][1]["type"]
            == GrapheneDynamicPartitionsRequestType.DELETE_PARTITIONS
        )

    def test_dry_run_failure(self, graphql_context: WorkspaceRequestContext):
        instigator_selector = infer_sensor_selector(graphql_context, "always_error_sensor")
        result = execute_dagster_graphql(
            graphql_context,
            SENSOR_DRY_RUN_MUTATION,
            variables={"selectorData": instigator_selector, "cursor": "blah"},
        )
        assert result.data
        assert result.data["sensorDryRun"]["__typename"] == "DryRunInstigationTick"
        evaluation_result = result.data["sensorDryRun"]["evaluationResult"]
        assert not evaluation_result["runRequests"]
        assert not evaluation_result["skipReason"]
        assert evaluation_result["dynamicPartitionsRequests"] is None
        assert (
            "Error occurred during the execution of evaluation_fn"
            in evaluation_result["error"]["message"]
        )

    def test_dry_run_skip(self, graphql_context: WorkspaceRequestContext):
        instigator_selector = infer_sensor_selector(graphql_context, "never_no_config_sensor")
        result = execute_dagster_graphql(
            graphql_context,
            SENSOR_DRY_RUN_MUTATION,
            variables={"selectorData": instigator_selector, "cursor": "blah"},
        )
        assert result.data
        assert result.data["sensorDryRun"]["__typename"] == "DryRunInstigationTick"
        evaluation_result = result.data["sensorDryRun"]["evaluationResult"]
        assert not evaluation_result["runRequests"]
        assert evaluation_result["skipReason"] == "never"
        assert not evaluation_result["error"]

    def test_dry_run_non_existent_sensor(self, graphql_context: WorkspaceRequestContext):
        unknown_instigator_selector = infer_sensor_selector(graphql_context, "sensor_doesnt_exist")
        with pytest.raises(UserFacingGraphQLError, match="GrapheneSensorNotFoundError"):
            execute_dagster_graphql(
                graphql_context,
                SENSOR_DRY_RUN_MUTATION,
                variables={"selectorData": unknown_instigator_selector, "cursor": "blah"},
            )
        unknown_repo_selector = {**unknown_instigator_selector}
        unknown_repo_selector["repositoryName"] = "doesnt_exist"
        with pytest.raises(UserFacingGraphQLError, match="GrapheneSensorNotFoundError"):
            execute_dagster_graphql(
                graphql_context,
                SENSOR_DRY_RUN_MUTATION,
                variables={"selectorData": unknown_repo_selector, "cursor": "blah"},
            )

        unknown_repo_location_selector = {**unknown_instigator_selector}
        unknown_repo_location_selector["repositoryLocationName"] = "doesnt_exist"
        with pytest.raises(UserFacingGraphQLError, match="GrapheneRepositoryLocationNotFound"):
            execute_dagster_graphql(
                graphql_context,
                SENSOR_DRY_RUN_MUTATION,
                variables={"selectorData": unknown_repo_location_selector, "cursor": "blah"},
            )

    def test_dry_run_cursor_updates(self, graphql_context: WorkspaceRequestContext):
        # Ensure that cursor does not update between dry runs
        selector = infer_sensor_selector(graphql_context, "update_cursor_sensor")
        result = execute_dagster_graphql(
            graphql_context,
            SENSOR_DRY_RUN_MUTATION,
            variables={"selectorData": selector, "cursor": None},
        )
        assert result.data
        assert result.data["sensorDryRun"]["__typename"] == "DryRunInstigationTick"
        evaluation_result = result.data["sensorDryRun"]["evaluationResult"]
        assert evaluation_result["cursor"] == "1"
        result = execute_dagster_graphql(
            graphql_context,
            GET_SENSOR_CURSOR_QUERY,
            variables={"sensorSelector": selector},
        )
        assert result.data
        assert result.data["sensorOrError"]["__typename"] == "Sensor"
        sensor = result.data["sensorOrError"]
        cursor = sensor["sensorState"]["typeSpecificData"]["lastCursor"]
        assert not cursor

    def test_get_sensors(self, graphql_context: WorkspaceRequestContext, snapshot):
        selector = infer_repository_selector(graphql_context)
        result = execute_dagster_graphql(
            graphql_context,
            GET_SENSORS_QUERY,
            variables={"repositorySelector": selector},
        )

        assert result.data
        assert result.data["sensorsOrError"]
        assert result.data["sensorsOrError"]["__typename"] == "Sensors"
        results = result.data["sensorsOrError"]["results"]

        # Snapshot is different for test_dict_repo because it does not contain any asset jobs,
        # so the sensor targets for sensors with asset selections differ
        if selector["repositoryName"] != "test_dict_repo":
            snapshot.assert_match(results)

    def test_get_sensors_filtered(self, graphql_context: WorkspaceRequestContext, snapshot):
        selector = infer_repository_selector(graphql_context)
        result = execute_dagster_graphql(
            graphql_context,
            GET_SENSORS_BY_STATUS_QUERY,
            variables={"repositorySelector": selector, "status": "RUNNING"},
        )

        assert result.data
        assert result.data["sensorsOrError"]
        assert result.data["sensorsOrError"]["__typename"] == "Sensors"
        results = result.data["sensorsOrError"]["results"]
        snapshot.assert_match(results)
        # running status includes automatically running sensors
        assert "running_in_code_sensor" in {
            sensor["name"] for sensor in result.data["sensorsOrError"]["results"]
        }

        result = execute_dagster_graphql(
            graphql_context,
            GET_SENSORS_BY_STATUS_QUERY,
            variables={"repositorySelector": selector, "status": "STOPPED"},
        )
        assert result.data
        assert result.data["sensorsOrError"]
        assert result.data["sensorsOrError"]["__typename"] == "Sensors"
        results = result.data["sensorsOrError"]["results"]
        assert "running_in_code_sensor" not in {
            sensor["name"] for sensor in result.data["sensorsOrError"]["results"]
        }

    def test_get_sensor(self, graphql_context: WorkspaceRequestContext, snapshot):
        sensor_selector = infer_sensor_selector(
            graphql_context, "always_no_config_sensor_with_tags_and_metadata"
        )
        result = execute_dagster_graphql(
            graphql_context,
            GET_SENSOR_QUERY,
            variables={"sensorSelector": sensor_selector},
        )

        assert result.data
        assert result.data["sensorOrError"]
        assert result.data["sensorOrError"]["__typename"] == "Sensor"
        sensor = result.data["sensorOrError"]
        snapshot.assert_match(sensor)
        assert sensor["sensorType"] == "STANDARD"
        assert sensor["tags"] == [{"key": "foo", "value": "bar"}]
        assert sensor["metadataEntries"] == [{"label": "foo", "text": "bar"}]

    def test_sensor_owners(self, graphql_context: WorkspaceRequestContext):
        sensor_selector = infer_sensor_selector(graphql_context, "owned_sensor")
        result = execute_dagster_graphql(
            graphql_context,
            GET_SENSOR_QUERY,
            variables={"sensorSelector": sensor_selector},
        )
        assert result.data
        assert result.data["sensorOrError"]
        assert result.data["sensorOrError"]["__typename"] == "Sensor"
        sensor = result.data["sensorOrError"]

        assert sensor["owners"] is not None
        assert len(sensor["owners"]) == 2

        # Check the user owner
        user_owner = None
        team_owner = None
        for owner in sensor["owners"]:
            if owner.get("email"):
                user_owner = owner
            elif owner.get("team"):
                team_owner = owner

        assert user_owner is not None
        assert user_owner["email"] == "test@elementl.com"
        assert team_owner is not None
        assert team_owner["team"] == "foo"


class TestReadonlySensorPermissions(ReadonlyGraphQLContextTestMatrix):
    def test_start_sensor_failure(self, graphql_context: WorkspaceRequestContext):
        sensor_selector = infer_sensor_selector(
            graphql_context, "always_no_config_sensor_with_tags_and_metadata"
        )
        result = execute_dagster_graphql(
            graphql_context,
            START_SENSORS_QUERY,
            variables={"sensorSelector": sensor_selector},
        )
        assert result.data

        assert result.data["startSensor"]["__typename"] == "UnauthorizedError"

    def test_stop_sensor_failure(self, graphql_context: WorkspaceRequestContext):
        sensor_selector = infer_sensor_selector(
            graphql_context, "always_no_config_sensor_with_tags_and_metadata"
        )

        result = execute_dagster_graphql(
            graphql_context,
            GET_SENSOR_STATUS_QUERY,
            variables={"sensorSelector": sensor_selector},
        )

        assert result.data["sensorOrError"]["sensorState"]["hasStartPermission"] is False
        assert result.data["sensorOrError"]["sensorState"]["hasStopPermission"] is False

        sensor_id = result.data["sensorOrError"]["sensorState"]["id"]

        stop_result = execute_dagster_graphql(
            graphql_context,
            STOP_SENSORS_QUERY,
            variables={
                "id": sensor_id,
            },
        )

        assert stop_result.data["stopSensor"]["__typename"] == "UnauthorizedError"

    def test_set_cursor_failure(self, graphql_context: WorkspaceRequestContext):
        selector = infer_sensor_selector(
            graphql_context, "always_no_config_sensor_with_tags_and_metadata"
        )

        result = execute_dagster_graphql(
            graphql_context,
            SET_SENSOR_CURSOR_MUTATION,
            variables={"sensorSelector": selector, "cursor": "foo"},
        )
        assert result.data
        assert result.data["setSensorCursor"]["__typename"] == "UnauthorizedError"


class TestSensorMutations(ExecutingGraphQLContextTestMatrix):
    def test_start_sensor(self, graphql_context: WorkspaceRequestContext):
        sensor_selector = infer_sensor_selector(
            graphql_context, "always_no_config_sensor_with_tags_and_metadata"
        )
        result = execute_dagster_graphql(
            graphql_context,
            START_SENSORS_QUERY,
            variables={"sensorSelector": sensor_selector},
        )
        assert result.data

        assert result.data["startSensor"]["sensorState"]["status"] == InstigatorStatus.RUNNING.value

    def test_stop_sensor(self, graphql_context: WorkspaceRequestContext):
        sensor_selector = infer_sensor_selector(
            graphql_context, "always_no_config_sensor_with_tags_and_metadata"
        )

        # start sensor
        start_result = execute_dagster_graphql(
            graphql_context,
            START_SENSORS_QUERY,
            variables={"sensorSelector": sensor_selector},
        )
        assert (
            start_result.data["startSensor"]["sensorState"]["status"]
            == InstigatorStatus.RUNNING.value
        )

        sensor_id = start_result.data["startSensor"]["id"]

        result = execute_dagster_graphql(
            graphql_context,
            STOP_SENSORS_QUERY,
            variables={"id": sensor_id},
        )
        assert result.data
        assert (
            result.data["stopSensor"]["instigationState"]["status"]
            == InstigatorStatus.STOPPED.value
        )

    def test_set_cursor(self, graphql_context: WorkspaceRequestContext):
        def get_cursor(selector):
            result = execute_dagster_graphql(
                graphql_context,
                GET_SENSOR_CURSOR_QUERY,
                variables={"sensorSelector": selector},
            )
            assert result.data
            assert result.data["sensorOrError"]["__typename"] == "Sensor"
            sensor = result.data["sensorOrError"]
            return sensor["sensorState"]["typeSpecificData"]["lastCursor"]

        def set_cursor(selector, cursor):
            result = execute_dagster_graphql(
                graphql_context,
                SET_SENSOR_CURSOR_MUTATION,
                variables={"sensorSelector": selector, "cursor": cursor},
            )
            assert result.data
            assert result.data["setSensorCursor"]["__typename"] == "Sensor"
            sensor = result.data["setSensorCursor"]
            return sensor["sensorState"]["typeSpecificData"]["lastCursor"]

        sensor_selector = infer_sensor_selector(
            graphql_context, "always_no_config_sensor_with_tags_and_metadata"
        )
        assert get_cursor(sensor_selector) is None
        set_cursor(sensor_selector, "new cursor value")
        assert get_cursor(sensor_selector) == "new cursor value"
        set_cursor(sensor_selector, None)
        assert get_cursor(sensor_selector) is None

        sensor_selector = infer_sensor_selector(graphql_context, "running_in_code_sensor")
        assert get_cursor(sensor_selector) is None
        set_cursor(sensor_selector, "new cursor value")
        assert get_cursor(sensor_selector) == "new cursor value"
        set_cursor(sensor_selector, None)
        assert get_cursor(sensor_selector) is None

    def test_start_sensor_with_default_status(self, graphql_context: WorkspaceRequestContext):
        sensor_selector = infer_sensor_selector(graphql_context, "running_in_code_sensor")

        result = execute_dagster_graphql(
            graphql_context,
            GET_SENSOR_STATUS_QUERY,
            variables={"sensorSelector": sensor_selector},
        )

        assert result.data["sensorOrError"]["defaultStatus"] == "RUNNING"
        assert result.data["sensorOrError"]["sensorState"]["status"] == "RUNNING"
        sensor_id = result.data["sensorOrError"]["sensorState"]["id"]

        assert result.data["sensorOrError"]["sensorState"]["hasStartPermission"] is True
        assert result.data["sensorOrError"]["sensorState"]["hasStopPermission"] is True

        start_result = execute_dagster_graphql(
            graphql_context,
            START_SENSORS_QUERY,
            variables={"sensorSelector": sensor_selector},
        )

        assert start_result.data["startSensor"]["sensorState"]["status"] == "RUNNING"

        stop_result = execute_dagster_graphql(
            graphql_context,
            STOP_SENSORS_QUERY,
            variables={
                "id": sensor_id,
            },
        )

        assert stop_result.data["stopSensor"]["instigationState"]["status"] == "STOPPED"

        # Now can be manually started
        start_result = execute_dagster_graphql(
            graphql_context,
            START_SENSORS_QUERY,
            variables={"sensorSelector": sensor_selector},
        )

        cid = CompoundID.from_string(sensor_id)
        instigator_state = graphql_context.instance.get_instigator_state(
            cid.remote_origin_id, cid.selector_id
        )

        assert instigator_state
        assert instigator_state.status == InstigatorStatus.RUNNING
        assert start_result.data["startSensor"]["sensorState"]["status"] == "RUNNING"

    def test_reset_sensor(self, graphql_context: WorkspaceRequestContext):
        sensor_selector = infer_sensor_selector(
            graphql_context, "always_no_config_sensor_with_tags_and_metadata"
        )
        result = execute_dagster_graphql(
            graphql_context,
            GET_SENSOR_STATUS_QUERY,
            variables={"sensorSelector": sensor_selector},
        )

        assert result.data["sensorOrError"]["defaultStatus"] == "STOPPED"
        assert result.data["sensorOrError"]["canReset"] is False
        assert result.data["sensorOrError"]["sensorState"]["status"] == "STOPPED"

        sensor_id = result.data["sensorOrError"]["sensorState"]["id"]

        start_result = execute_dagster_graphql(
            graphql_context,
            START_SENSORS_QUERY,
            variables={"sensorSelector": sensor_selector},
        )

        assert start_result.data["startSensor"]["canReset"] is True
        assert start_result.data["startSensor"]["sensorState"]["status"] == "RUNNING"

        # Resetting a sensor that is already running stops it
        reset_result = execute_dagster_graphql(
            graphql_context,
            RESET_SENSORS_QUERY,
            variables={"sensorSelector": sensor_selector},
        )
        cid = CompoundID.from_string(sensor_id)
        instigator_state = graphql_context.instance.get_instigator_state(
            cid.remote_origin_id, cid.selector_id
        )

        assert instigator_state
        assert instigator_state.status == InstigatorStatus.DECLARED_IN_CODE
        assert reset_result.data["resetSensor"]["canReset"] is False
        assert reset_result.data["resetSensor"]["sensorState"]["status"] == "STOPPED"

        # Resetting a stopped sensor is a noop
        reset_result = execute_dagster_graphql(
            graphql_context,
            RESET_SENSORS_QUERY,
            variables={"sensorSelector": sensor_selector},
        )

        instigator_state = graphql_context.instance.get_instigator_state(
            cid.remote_origin_id, cid.selector_id
        )

        assert instigator_state
        assert instigator_state.status == InstigatorStatus.DECLARED_IN_CODE
        assert reset_result.data["resetSensor"]["canReset"] is False
        assert reset_result.data["resetSensor"]["sensorState"]["status"] == "STOPPED"

    def test_reset_sensor_with_default_status(self, graphql_context: WorkspaceRequestContext):
        sensor_selector = infer_sensor_selector(graphql_context, "running_in_code_sensor")
        result = execute_dagster_graphql(
            graphql_context,
            GET_SENSOR_STATUS_QUERY,
            variables={"sensorSelector": sensor_selector},
        )

        assert result.data["sensorOrError"]["defaultStatus"] == "RUNNING"
        assert result.data["sensorOrError"]["canReset"] is False
        assert result.data["sensorOrError"]["sensorState"]["status"] == "RUNNING"
        assert result.data["sensorOrError"]["sensorState"]["hasStartPermission"] is True
        assert result.data["sensorOrError"]["sensorState"]["hasStopPermission"] is True

        sensor_id = result.data["sensorOrError"]["sensorState"]["id"]

        stop_result = execute_dagster_graphql(
            graphql_context,
            STOP_SENSORS_QUERY,
            variables={"id": sensor_id},
        )

        assert stop_result.data["stopSensor"]["instigationState"]["status"] == "STOPPED"

        result = execute_dagster_graphql(
            graphql_context,
            GET_SENSOR_STATUS_QUERY,
            variables={"sensorSelector": sensor_selector},
        )

        assert result.data["sensorOrError"]["canReset"] is True

        # Now can be restarted
        start_result = execute_dagster_graphql(
            graphql_context,
            RESET_SENSORS_QUERY,
            variables={"sensorSelector": sensor_selector},
        )

        cid = CompoundID.from_string(sensor_id)
        instigator_state = graphql_context.instance.get_instigator_state(
            cid.remote_origin_id, cid.selector_id
        )

        assert instigator_state
        assert instigator_state.status == InstigatorStatus.DECLARED_IN_CODE
        assert start_result.data["resetSensor"]["canReset"] is False
        assert start_result.data["resetSensor"]["sensorState"]["status"] == "RUNNING"

    def test_sensor_with_default_status_stopped(self, graphql_context: WorkspaceRequestContext):
        sensor_selector = infer_sensor_selector(graphql_context, "stopped_in_code_sensor")
        result = execute_dagster_graphql(
            graphql_context,
            GET_SENSOR_STATUS_QUERY,
            variables={"sensorSelector": sensor_selector},
        )

        assert result.data["sensorOrError"]["defaultStatus"] == "STOPPED"
        assert result.data["sensorOrError"]["canReset"] is False
        assert result.data["sensorOrError"]["sensorState"]["status"] == "STOPPED"
        assert result.data["sensorOrError"]["sensorState"]["hasStartPermission"] is True
        assert result.data["sensorOrError"]["sensorState"]["hasStopPermission"] is True

        sensor_id = result.data["sensorOrError"]["sensorState"]["id"]
        instigation_selector = {
            "name": sensor_selector["sensorName"],
            "repositoryLocationName": sensor_selector["repositoryLocationName"],
            "repositoryName": sensor_selector["repositoryName"],
        }
        state_result = execute_dagster_graphql(
            graphql_context,
            INSTIGATION_STATE_QUERY,
            variables={
                "instigationSelector": instigation_selector,
                "id": sensor_id,
            },
        )
        assert state_result.data["instigationStateOrError"]["__typename"] == "InstigationState"
        assert state_result.data["instigationStateOrError"]["status"] == "STOPPED"

        stop_result = execute_dagster_graphql(
            graphql_context,
            START_SENSORS_QUERY,
            variables={"sensorSelector": sensor_selector},
        )

        assert stop_result.data["startSensor"]["sensorState"]["status"] == "RUNNING"

        result = execute_dagster_graphql(
            graphql_context,
            GET_SENSOR_STATUS_QUERY,
            variables={"sensorSelector": sensor_selector},
        )

        assert result.data["sensorOrError"]["canReset"] is True

        state_result = execute_dagster_graphql(
            graphql_context,
            INSTIGATION_STATE_QUERY,
            variables={
                "instigationSelector": instigation_selector,
                "id": sensor_id,
            },
        )
        assert state_result.data["instigationStateOrError"]["__typename"] == "InstigationState"
        assert state_result.data["instigationStateOrError"]["status"] == "RUNNING"

        # Now can be restarted
        start_result = execute_dagster_graphql(
            graphql_context,
            RESET_SENSORS_QUERY,
            variables={"sensorSelector": sensor_selector},
        )

        cid = CompoundID.from_string(sensor_id)
        instigator_state = graphql_context.instance.get_instigator_state(
            cid.remote_origin_id, cid.selector_id
        )

        assert instigator_state
        assert instigator_state.status == InstigatorStatus.DECLARED_IN_CODE
        assert start_result.data["resetSensor"]["canReset"] is False
        assert start_result.data["resetSensor"]["sensorState"]["status"] == "STOPPED"

        state_result = execute_dagster_graphql(
            graphql_context,
            INSTIGATION_STATE_QUERY,
            variables={
                "instigationSelector": instigation_selector,
                "id": sensor_id,
            },
        )
        assert state_result.data["instigationStateOrError"]["__typename"] == "InstigationState"
        assert state_result.data["instigationStateOrError"]["status"] == "STOPPED"


def test_sensor_next_ticks(graphql_context: WorkspaceRequestContext):
    remote_repository = graphql_context.get_code_location(main_repo_location_name()).get_repository(
        main_repo_name()
    )

    sensor_name = "always_no_config_sensor_with_tags_and_metadata"
    sensor = remote_repository.get_sensor(sensor_name)
    sensor_selector = infer_sensor_selector(graphql_context, sensor_name)

    result = execute_dagster_graphql(
        graphql_context, GET_SENSOR_QUERY, variables={"sensorSelector": sensor_selector}
    )

    # test default sensor off
    assert result.data
    assert result.data["sensorOrError"]["__typename"] == "Sensor"
    next_tick = result.data["sensorOrError"]["nextTick"]
    assert not next_tick

    error_sensor_name = "always_error_sensor"
    error_sensor = remote_repository.get_sensor(error_sensor_name)
    error_sensor_selector = infer_sensor_selector(graphql_context, error_sensor_name)

    # test default sensor with no tick
    graphql_context.instance.add_instigator_state(
        InstigatorState(sensor.get_remote_origin(), InstigatorType.SENSOR, InstigatorStatus.RUNNING)
    )
    graphql_context.instance.add_instigator_state(
        InstigatorState(
            error_sensor.get_remote_origin(),
            InstigatorType.SENSOR,
            InstigatorStatus.RUNNING,
        )
    )

    result = execute_dagster_graphql(
        graphql_context, GET_SENSOR_QUERY, variables={"sensorSelector": sensor_selector}
    )
    assert result.data
    assert len(result.data["sensorOrError"]["sensorState"]["ticks"]) == 0
    assert result.data["sensorOrError"]["__typename"] == "Sensor"
    next_tick = result.data["sensorOrError"]["nextTick"]
    assert not next_tick

    # test default sensor with last tick
    _create_tick(graphql_context)

    result = execute_dagster_graphql(
        graphql_context, GET_SENSOR_QUERY, variables={"sensorSelector": sensor_selector}
    )
    assert result.data
    assert result.data["sensorOrError"]["__typename"] == "Sensor"

    assert len(result.data["sensorOrError"]["sensorState"]["ticks"]) == 1
    assert not result.data["sensorOrError"]["sensorState"]["ticks"][0].get("error")

    next_tick = result.data["sensorOrError"]["nextTick"]
    assert next_tick

    error_result = execute_dagster_graphql(
        graphql_context, GET_SENSOR_QUERY, variables={"sensorSelector": error_sensor_selector}
    )
    assert error_result.data
    assert error_result.data["sensorOrError"]["__typename"] == "Sensor"
    assert len(error_result.data["sensorOrError"]["sensorState"]["ticks"]) == 1
    assert error_result.data["sensorOrError"]["sensorState"]["ticks"][0]["error"]


def _create_tick(graphql_context):
    logger = get_default_daemon_logger("SensorDaemon")
    futures = {}
    list(
        execute_sensor_iteration(
            graphql_context.process_context,
            logger,
            threadpool_executor=SingleThreadPoolExecutor(),
            submit_threadpool_executor=None,
            sensor_tick_futures=futures,
        )
    )
    wait_for_futures(futures)


def test_sensor_tick_range(graphql_context: WorkspaceRequestContext):
    remote_repository = graphql_context.get_code_location(main_repo_location_name()).get_repository(
        main_repo_name()
    )

    sensor_name = "always_no_config_sensor_with_tags_and_metadata"
    sensor = remote_repository.get_sensor(sensor_name)
    sensor_selector = infer_sensor_selector(graphql_context, sensor_name)

    # test with no job state
    result = execute_dagster_graphql(
        graphql_context,
        GET_SENSOR_TICK_RANGE_QUERY,
        variables={"sensorSelector": sensor_selector, "dayRange": None, "dayOffset": None},
    )
    assert len(result.data["sensorOrError"]["sensorState"]["ticks"]) == 0

    # turn the sensor on
    graphql_context.instance.add_instigator_state(
        InstigatorState(sensor.get_remote_origin(), InstigatorType.SENSOR, InstigatorStatus.RUNNING)
    )

    now = datetime.datetime.now(tz=get_timezone("US/Central"))
    one = now - datetime.timedelta(days=2) - datetime.timedelta(hours=1)
    with freeze_time(one):
        _create_tick(graphql_context)

    two = now - relativedelta(days=1) - datetime.timedelta(hours=1)
    with freeze_time(two):
        _create_tick(graphql_context)

    three = now - datetime.timedelta(hours=1)
    with freeze_time(three):
        _create_tick(graphql_context)

    result = execute_dagster_graphql(
        graphql_context,
        GET_SENSOR_TICK_RANGE_QUERY,
        variables={"sensorSelector": sensor_selector, "dayRange": None, "dayOffset": None},
    )
    assert len(result.data["sensorOrError"]["sensorState"]["ticks"]) == 3

    result = execute_dagster_graphql(
        graphql_context,
        GET_SENSOR_TICK_RANGE_QUERY,
        variables={"sensorSelector": sensor_selector, "dayRange": 1, "dayOffset": None},
    )
    assert len(result.data["sensorOrError"]["sensorState"]["ticks"]) == 1
    assert result.data["sensorOrError"]["sensorState"]["ticks"][0]["timestamp"] == three.timestamp()
    assert (
        result.data["sensorOrError"]["sensorState"]["ticks"][0]["endTimestamp"] >= three.timestamp()
    )

    result = execute_dagster_graphql(
        graphql_context,
        GET_SENSOR_TICK_RANGE_QUERY,
        variables={
            "sensorSelector": sensor_selector,
            "beforeTimestamp": three.timestamp() + 1,
            "afterTimestamp": three.timestamp() - 1,
        },
    )
    assert len(result.data["sensorOrError"]["sensorState"]["ticks"]) == 1
    assert result.data["sensorOrError"]["sensorState"]["ticks"][0]["timestamp"] == three.timestamp()
    assert (
        result.data["sensorOrError"]["sensorState"]["ticks"][0]["endTimestamp"] >= three.timestamp()
    )

    result = execute_dagster_graphql(
        graphql_context,
        GET_SENSOR_TICK_RANGE_QUERY,
        variables={"sensorSelector": sensor_selector, "dayRange": 1, "dayOffset": 1},
    )
    assert len(result.data["sensorOrError"]["sensorState"]["ticks"]) == 1
    assert result.data["sensorOrError"]["sensorState"]["ticks"][0]["timestamp"] == two.timestamp()
    assert (
        result.data["sensorOrError"]["sensorState"]["ticks"][0]["endTimestamp"] >= two.timestamp()
    )

    result = execute_dagster_graphql(
        graphql_context,
        GET_SENSOR_TICK_RANGE_QUERY,
        variables={
            "sensorSelector": sensor_selector,
            "dayRange": 2,
            "dayOffset": None,
        },
    )
    assert len(result.data["sensorOrError"]["sensorState"]["ticks"]) == 2


def test_sensor_type_query(graphql_context: WorkspaceRequestContext):
    instance = graphql_context.instance
    if not instance.supports_batch_tick_queries:
        pytest.skip("storage cannot batch fetch")

    selector = infer_repository_selector(graphql_context)
    result = execute_dagster_graphql(
        graphql_context,
        REPOSITORY_SENSORS_QUERY,
        variables={"repositorySelector": selector, "sensorType": SensorType.AUTO_MATERIALIZE.value},
    )
    assert len(result.data["repositoryOrError"]["sensors"]) == 1
    assert result.data["repositoryOrError"]["sensors"][0]["name"] == "my_auto_materialize_sensor"


def test_repository_batching(graphql_context: WorkspaceRequestContext):
    instance = graphql_context.instance
    if not instance.supports_batch_tick_queries:
        pytest.skip("storage cannot batch fetch")

    counter = Counter()
    traced_counter.set(counter)
    selector = infer_repository_selector(graphql_context)
    result = execute_dagster_graphql(
        graphql_context,
        REPOSITORY_SENSORS_QUERY,
        variables={"repositorySelector": selector},
    )
    assert result.data
    assert "repositoryOrError" in result.data
    assert "sensors" in result.data["repositoryOrError"]
    counts = counter.counts()
    assert counts
    assert len(counts) == 3

    # We should have a single batch call to fetch instigator state, instead of separate calls for
    # each sensor (~5 distinct sensors in the repo)
    # 1) `get_batch_ticks` is called to fetch all the ticks for the sensors
    # 2) `all_instigator_state` is fetched to instantiate GrapheneSensor
    assert counts.get("SchedulingMethods.get_batch_ticks") == 1
    assert counts.get("SchedulingMethods.all_instigator_state") == 1


def test_sensor_ticks_filtered(graphql_context: WorkspaceRequestContext):
    remote_repository = graphql_context.get_code_location(main_repo_location_name()).get_repository(
        main_repo_name()
    )

    repository_selector = {
        "repositoryLocationName": main_repo_location_name(),
        "repositoryName": main_repo_name(),
    }
    sensor_name = "always_no_config_sensor_with_tags_and_metadata"
    sensor = remote_repository.get_sensor(sensor_name)
    sensor_selector = infer_sensor_selector(graphql_context, sensor_name)

    # turn the sensor on
    graphql_context.instance.add_instigator_state(
        InstigatorState(sensor.get_remote_origin(), InstigatorType.SENSOR, InstigatorStatus.RUNNING)
    )

    now = datetime.datetime.now(tz=get_timezone("US/Central"))
    with freeze_time(now):
        _create_tick(graphql_context)  # create a success tick

    success_tick_id = graphql_context.instance.get_ticks(
        origin_id=sensor.get_remote_origin().get_id(),
        selector_id=sensor.get_remote_origin().get_selector().get_id(),
        limit=1,
    )[0].tick_id

    # create a started tick
    started_tick_id, _ = graphql_context.instance.create_tick(
        TickData(
            instigator_origin_id=sensor.get_remote_origin().get_id(),
            instigator_name=sensor_name,
            instigator_type=InstigatorType.SENSOR,
            status=TickStatus.STARTED,
            timestamp=now.timestamp(),
            selector_id=sensor.selector_id,
        )
    )

    # create a skipped tick
    skipped_tick_id, _ = graphql_context.instance.create_tick(
        TickData(
            instigator_origin_id=sensor.get_remote_origin().get_id(),
            instigator_name=sensor_name,
            instigator_type=InstigatorType.SENSOR,
            status=TickStatus.SKIPPED,
            timestamp=now.timestamp(),
            selector_id=sensor.selector_id,
        )
    )

    # create a failed tick
    failed_tick_id, _ = graphql_context.instance.create_tick(
        TickData(
            instigator_origin_id=sensor.get_remote_origin().get_id(),
            instigator_name=sensor_name,
            instigator_type=InstigatorType.SENSOR,
            status=TickStatus.FAILURE,
            timestamp=now.timestamp(),
            error=SerializableErrorInfo(message="foobar", stack=[], cls_name=None, cause=None),
            selector_id=sensor.selector_id,
        )
    )

    result = execute_dagster_graphql(
        graphql_context,
        GET_TICKS_QUERY,
        variables={"sensorSelector": sensor_selector, "tickId": str(started_tick_id)},
    )
    assert len(result.data["sensorOrError"]["sensorState"]["ticks"]) == 4

    result = execute_dagster_graphql(
        graphql_context,
        GET_TICKS_QUERY,
        variables={
            "sensorSelector": sensor_selector,
            "statuses": ["STARTED"],
            "tickId": str(started_tick_id),
        },
    )
    assert len(result.data["sensorOrError"]["sensorState"]["ticks"]) == 1
    assert result.data["sensorOrError"]["sensorState"]["ticks"][0]["status"] == "STARTED"
    assert (
        result.data["sensorOrError"]["sensorState"]["ticks"][0]["tickId"]
        == result.data["sensorOrError"]["sensorState"]["tick"]["tickId"]
        == str(started_tick_id)
    )

    result = execute_dagster_graphql(
        graphql_context,
        GET_TICKS_QUERY,
        variables={
            "sensorSelector": sensor_selector,
            "statuses": ["FAILURE"],
            "tickId": str(failed_tick_id),
        },
    )
    assert len(result.data["sensorOrError"]["sensorState"]["ticks"]) == 1
    assert result.data["sensorOrError"]["sensorState"]["ticks"][0]["status"] == "FAILURE"
    assert (
        result.data["sensorOrError"]["sensorState"]["ticks"][0]["tickId"]
        == result.data["sensorOrError"]["sensorState"]["tick"]["tickId"]
        == str(failed_tick_id)
    )

    result = execute_dagster_graphql(
        graphql_context,
        GET_TICKS_QUERY,
        variables={
            "sensorSelector": sensor_selector,
            "statuses": ["SKIPPED"],
            "tickId": str(skipped_tick_id),
        },
    )
    assert len(result.data["sensorOrError"]["sensorState"]["ticks"]) == 1
    assert result.data["sensorOrError"]["sensorState"]["ticks"][0]["status"] == "SKIPPED"
    assert (
        result.data["sensorOrError"]["sensorState"]["ticks"][0]["tickId"]
        == result.data["sensorOrError"]["sensorState"]["tick"]["tickId"]
        == str(skipped_tick_id)
    )

    result = execute_dagster_graphql(
        graphql_context,
        GET_SENSORS_TICKS_BY_STATUS_QUERY,
        variables={
            "repositorySelector": repository_selector,
            "statuses": ["FAILURE", "SUCCESS"],
        },
    )
    all_sensors = result.data["sensorsOrError"]["results"]
    selected_sensor = next(
        (sensor for sensor in all_sensors if sensor["name"] == sensor_selector["sensorName"]), None
    )
    assert selected_sensor
    assert len(selected_sensor["sensorState"]["ticks"]) == 2

    backfill_id = make_new_backfill_id()
    graphql_context.instance.create_tick(
        TickData(
            instigator_origin_id=sensor.get_remote_origin().get_id(),
            instigator_name=sensor_name,
            instigator_type=InstigatorType.SENSOR,
            status=TickStatus.SUCCESS,
            timestamp=now.timestamp(),
            selector_id=sensor.selector_id,
            origin_run_ids=[backfill_id],
            run_ids=[backfill_id],
        )
    )

    result = execute_dagster_graphql(
        graphql_context,
        GET_TICKS_QUERY,
        variables={
            "sensorSelector": sensor_selector,
            "statuses": ["SUCCESS"],
            "tickId": str(success_tick_id),
        },
    )
    assert len(result.data["sensorOrError"]["sensorState"]["ticks"]) == 2
    assert any(
        len(tick["runs"]) > 0 for tick in result.data["sensorOrError"]["sensorState"]["ticks"]
    )


def _get_unloadable_sensor_origin(name):
    working_directory = os.path.dirname(__file__)
    loadable_target_origin = LoadableTargetOrigin(
        executable_path=sys.executable,
        python_file=__file__,
        working_directory=working_directory,
    )
    return RemoteRepositoryOrigin(
        InProcessCodeLocationOrigin(loadable_target_origin), "fake_repository"
    ).get_instigator_origin(name)


def test_unloadable_sensor(graphql_context: WorkspaceRequestContext):
    instance = graphql_context.instance

    running_origin = _get_unloadable_sensor_origin("unloadable_running")
    running_instigator_state = InstigatorState(
        running_origin,
        InstigatorType.SENSOR,
        InstigatorStatus.RUNNING,
        SensorInstigatorData(min_interval=30, cursor=None, sensor_type=SensorType.STANDARD),
    )

    stopped_origin = _get_unloadable_sensor_origin("unloadable_stopped")

    instance.add_instigator_state(running_instigator_state)

    instance.add_instigator_state(
        InstigatorState(
            stopped_origin,
            InstigatorType.SENSOR,
            InstigatorStatus.STOPPED,
            SensorInstigatorData(min_interval=30, cursor=None, sensor_type=SensorType.STANDARD),
        )
    )

    # Verify that we can stop the unloadable sensor
    stop_result = execute_dagster_graphql(
        graphql_context,
        STOP_SENSORS_QUERY,
        variables={
            "jobOriginId": running_instigator_state.instigator_origin_id,
            "jobSelectorId": running_instigator_state.selector_id,
        },
    )
    assert (
        stop_result.data["stopSensor"]["instigationState"]["status"]
        == InstigatorStatus.STOPPED.value
    )


def test_sensor_tick_logs(graphql_context: WorkspaceRequestContext):
    instance = graphql_context.instance
    remote_repository = graphql_context.get_code_location(main_repo_location_name()).get_repository(
        main_repo_name()
    )

    sensor_name = "logging_sensor"
    sensor = remote_repository.get_sensor(sensor_name)
    sensor_selector = infer_sensor_selector(graphql_context, sensor_name)

    # turn the sensor on
    instance.add_instigator_state(
        InstigatorState(sensor.get_remote_origin(), InstigatorType.SENSOR, InstigatorStatus.RUNNING)
    )

    _create_tick(graphql_context)

    result = execute_dagster_graphql(
        graphql_context,
        GET_TICK_LOGS_QUERY,
        variables={"sensorSelector": sensor_selector},
    )
    assert len(result.data["sensorOrError"]["sensorState"]["ticks"]) == 1
    tick = result.data["sensorOrError"]["sensorState"]["ticks"][0]
    log_messages = tick["logEvents"]["events"]
    assert len(log_messages) == 2
    assert log_messages[0]["message"] == "hello hello"
    assert log_messages[1]["message"].startswith("goodbye goodbye")
    assert "Traceback" in log_messages[1]["message"]
    assert "Exception: hi hi" in log_messages[1]["message"]


def test_sensor_dynamic_partitions_request_results(graphql_context: WorkspaceRequestContext):
    instance = graphql_context.instance
    remote_repository = graphql_context.get_code_location(main_repo_location_name()).get_repository(
        main_repo_name()
    )

    sensor_name = "dynamic_partition_requesting_sensor"
    sensor = remote_repository.get_sensor(sensor_name)
    sensor_selector = infer_sensor_selector(graphql_context, sensor_name)

    instance.add_dynamic_partitions("foo", ["existent_key", "old_key"])

    # turn the sensor on
    instance.add_instigator_state(
        InstigatorState(sensor.get_remote_origin(), InstigatorType.SENSOR, InstigatorStatus.RUNNING)
    )

    _create_tick(graphql_context)

    result = execute_dagster_graphql(
        graphql_context,
        GET_TICK_DYNAMIC_PARTITIONS_REQUEST_RESULTS_QUERY,
        variables={"sensorSelector": sensor_selector},
    )
    assert len(result.data["sensorOrError"]["sensorState"]["ticks"]) == 1
    tick = result.data["sensorOrError"]["sensorState"]["ticks"][0]
    results = tick["dynamicPartitionsRequestResults"]
    assert len(results) == 2
    assert results[0]["partitionsDefName"] == "foo"
    assert results[0]["type"] == "ADD_PARTITIONS"
    assert results[0]["partitionKeys"] == ["new_key", "new_key2"]
    assert results[0]["skippedPartitionKeys"] == ["existent_key"]

    assert results[1]["partitionsDefName"] == "foo"
    assert results[1]["type"] == "DELETE_PARTITIONS"
    assert results[1]["partitionKeys"] == ["old_key"]
    assert results[1]["skippedPartitionKeys"] == ["nonexistent_key"]


def test_asset_selection(graphql_context):
    sensor_name = "my_auto_materialize_sensor"
    sensor_selector = infer_sensor_selector(graphql_context, sensor_name)

    result = execute_dagster_graphql(
        graphql_context, GET_SENSOR_QUERY, variables={"sensorSelector": sensor_selector}
    )

    assert result.data
    assert result.data["sensorOrError"]["__typename"] == "Sensor"

    assert result.data["sensorOrError"]["assetSelection"]["assetSelectionString"] == (
        '(key:"fresh_diamond_bottom" or key:"asset_with_automation_condition"'
        ' or key:"asset_with_custom_automation_condition") and code_location:"test_location"'
    )
    assert result.data["sensorOrError"]["assetSelection"]["assetKeys"] == [
        {"path": ["asset_with_automation_condition"]},
        {"path": ["asset_with_custom_automation_condition"]},
        {"path": ["fresh_diamond_bottom"]},
    ]
    assert result.data["sensorOrError"]["assetSelection"]["assets"] == [
        {
            "key": {"path": ["asset_with_automation_condition"]},
            "definition": {"assetKey": {"path": ["asset_with_automation_condition"]}},
        },
        {
            "key": {"path": ["asset_with_custom_automation_condition"]},
            "definition": {"assetKey": {"path": ["asset_with_custom_automation_condition"]}},
        },
        {
            "key": {"path": ["fresh_diamond_bottom"]},
            "definition": {"assetKey": {"path": ["fresh_diamond_bottom"]}},
        },
    ]
    assert result.data["sensorOrError"]["assetSelection"]["assetsOrError"]["nodes"] == [
        {
            "key": {"path": ["asset_with_automation_condition"]},
        },
        {
            "key": {"path": ["asset_with_custom_automation_condition"]},
        },
        {
            "key": {"path": ["fresh_diamond_bottom"]},
        },
    ]


def test_jobless_asset_selection(graphql_context):
    sensor_name = "jobless_sensor"
    sensor_selector = infer_sensor_selector(graphql_context, sensor_name)

    result = execute_dagster_graphql(
        graphql_context, GET_SENSOR_QUERY, variables={"sensorSelector": sensor_selector}
    )

    assert result.data
    assert result.data["sensorOrError"]["__typename"] == "Sensor"
    assert (
        result.data["sensorOrError"]["assetSelection"]["assetSelectionString"]
        == 'key:"asset_one" and code_location:"test_location"'
    )
    assert result.data["sensorOrError"]["assetSelection"]["assetKeys"] == [{"path": ["asset_one"]}]
    assert result.data["sensorOrError"]["assetSelection"]["assets"] == [
        {
            "key": {"path": ["asset_one"]},
            "definition": {"assetKey": {"path": ["asset_one"]}},
        }
    ]
    assert result.data["sensorOrError"]["assetSelection"]["assetsOrError"]["nodes"] == [
        {
            "key": {"path": ["asset_one"]},
        }
    ]


def test_invalid_sensor_asset_selection(graphql_context):
    sensor_name = "invalid_asset_selection_error"
    sensor_selector = infer_sensor_selector(graphql_context, sensor_name)

    result = execute_dagster_graphql(
        graphql_context,
        GET_ASSET_SELECTION_ERROR_QUERY,
        variables={"sensorSelector": sensor_selector},
    )

    assert result.data
    assert result.data["sensorOrError"]["__typename"] == "Sensor"
    assert (
        result.data["sensorOrError"]["assetSelection"]["assetSelectionString"]
        == 'key:"does_not_exist" and code_location:"test_location"'
    )
    assert (
        result.data["sensorOrError"]["assetSelection"]["assetsOrError"]["__typename"]
        == "PythonError"
    )
    assert (
        "no AssetsDefinition objects supply these keys"
        in result.data["sensorOrError"]["assetSelection"]["assetsOrError"]["message"]
    )



import graphene

from dagster_graphql.implementation.execution import gen_captured_log_data, gen_events_for_run
from dagster_graphql.schema.external import (
    GrapheneLocationStateChangeSubscription,
    gen_location_state_changes,
)
from dagster_graphql.schema.logs.compute_logs import GrapheneCapturedLogs
from dagster_graphql.schema.pipelines.subscription import GraphenePipelineRunLogsSubscriptionPayload
from dagster_graphql.schema.util import ResolveInfo, non_null_list


class GrapheneSubscription(graphene.ObjectType):
    """The root for all subscriptions to retrieve real-time data from the Dagster instance."""

    class Meta:
        name = "Subscription"

    pipelineRunLogs = graphene.Field(
        graphene.NonNull(GraphenePipelineRunLogsSubscriptionPayload),
        runId=graphene.Argument(graphene.NonNull(graphene.ID)),
        cursor=graphene.Argument(
            graphene.String,
            description=(
                "A cursor retrieved from the API. Pass 'HEAD' to stream from the current event"
                " onward."
            ),
        ),
        description="Retrieve real-time event logs after applying a filter on run id and cursor.",
    )

    capturedLogs = graphene.Field(
        graphene.NonNull(GrapheneCapturedLogs),
        logKey=graphene.Argument(non_null_list(graphene.String)),
        cursor=graphene.Argument(graphene.String),
        description="Retrieve real-time compute logs.",
    )

    locationStateChangeEvents = graphene.Field(
        graphene.NonNull(GrapheneLocationStateChangeSubscription),
        description=(
            "Retrieve real-time events when a location in the workspace undergoes a state change."
        ),
    )

    def subscribe_pipelineRunLogs(self, graphene_info: ResolveInfo, runId, cursor=None):
        return gen_events_for_run(graphene_info, runId, cursor)

    def subscribe_capturedLogs(self, graphene_info: ResolveInfo, logKey, cursor=None):
        return gen_captured_log_data(graphene_info, logKey, cursor)

    def subscribe_locationStateChangeEvents(self, graphene_info: ResolveInfo):
        return gen_location_state_changes(graphene_info)



from collections.abc import Sequence
from typing import TYPE_CHECKING

import graphene
from dagster._core.definitions.asset_checks.asset_check_spec import AssetCheckKey
from dagster._core.definitions.asset_key import AssetKey
from dagster._core.definitions.asset_selection import AssetSelection, CodeLocationAssetSelection
from dagster._core.remote_representation.handle import RepositoryHandle

from dagster_graphql.implementation.fetch_assets import get_asset
from dagster_graphql.implementation.utils import capture_error
from dagster_graphql.schema.entity_key import GrapheneAssetCheckHandle, GrapheneAssetKey
from dagster_graphql.schema.util import ResolveInfo, non_null_list

if TYPE_CHECKING:
    from dagster_graphql.schema.roots.assets import GrapheneAssetConnection


class GrapheneAssetSelection(graphene.ObjectType):
    assetSelectionString = graphene.String()
    assetKeys = non_null_list(GrapheneAssetKey)
    assetChecks = non_null_list(GrapheneAssetCheckHandle)
    assets = non_null_list("dagster_graphql.schema.pipelines.pipeline.GrapheneAsset")
    assetsOrError = graphene.NonNull("dagster_graphql.schema.roots.assets.GrapheneAssetsOrError")

    def __init__(
        self,
        asset_selection: AssetSelection,
        repository_handle: RepositoryHandle,
    ):
        self._asset_selection = asset_selection
        self._repository_handle = repository_handle
        self._resolved_keys = None
        self._resolved_checks = None

    def resolve_assetSelectionString(self, _graphene_info) -> str:
        return str(
            self._asset_selection
            & CodeLocationAssetSelection(
                selected_code_location=self._repository_handle.code_location_origin.location_name
            )
        )

    def resolve_assetKeys(self, graphene_info: ResolveInfo):
        return [
            GrapheneAssetKey(path=asset_key.path)
            for asset_key in self._get_resolved_and_sorted_keys(graphene_info)
        ]

    def resolve_assetChecks(self, graphene_info: ResolveInfo):
        return [
            GrapheneAssetCheckHandle(handle)
            for handle in self._get_resolved_and_sorted_checks(graphene_info)
        ]

    def _get_assets(self, graphene_info: ResolveInfo):
        return [
            get_asset(asset_key) for asset_key in self._get_resolved_and_sorted_keys(graphene_info)
        ]

    def resolve_assets(self, graphene_info: ResolveInfo):
        return self._get_assets(graphene_info)

    def _get_resolved_and_sorted_keys(self, graphene_info: ResolveInfo) -> Sequence[AssetKey]:
        """Use this to maintain stability in ordering."""
        if self._resolved_keys is None:
            repo = graphene_info.context.get_repository(self._repository_handle)
            self._resolved_keys = sorted(self._asset_selection.resolve(repo.asset_graph), key=str)

        return self._resolved_keys

    def _get_resolved_and_sorted_checks(
        self, graphene_info: ResolveInfo
    ) -> Sequence[AssetCheckKey]:
        """Use this to maintain stability in ordering."""
        if self._resolved_checks is None:
            repo = graphene_info.context.get_repository(self._repository_handle)
            self._resolved_checks = sorted(
                self._asset_selection.resolve_checks(repo.asset_graph), key=str
            )

        return self._resolved_checks

    @capture_error
    def resolve_assetsOrError(self, graphene_info) -> "GrapheneAssetConnection":
        from dagster_graphql.schema.roots.assets import GrapheneAssetConnection

        return GrapheneAssetConnection(nodes=self._get_assets(graphene_info), cursor=None)

    class Meta:
        name = "AssetSelection"



from collections import defaultdict
from collections.abc import Sequence
from enum import Enum
from typing import Any

from dagster import (
    DagsterInstance,
    _check as check,
)
from dagster._core.remote_representation.external import RemoteRepository
from dagster._core.scheduler.instigation import InstigatorState, InstigatorType


class RepositoryDataType(Enum):
    JOB_RUNS = "job_runs"
    SCHEDULE_RUNS = "schedule_runs"
    SENSOR_RUNS = "sensor_runs"
    SCHEDULE_STATES = "schedule_states"
    SENSOR_STATES = "sensor_states"
    SCHEDULE_TICKS = "schedule_ticks"
    SENSOR_TICKS = "sensor_ticks"


class RepositoryScopedBatchLoader:
    """A batch loader that fetches an assortment of data for a given repository.  This loader is
    expected to be instantiated once per repository, and then passed to various child graphene
    objects to batch calls to the DB.

    We can instantiate this loader without knowing which data we will end up requesting deeper in
    the graphql nested schema, which means we can batch DB requests without changing the structure
     of our graphql request.

    Example: When the last 10 runs are requested for a job in the repository, we know that they will
    be fetched for every job in the repository.  We can batch fetch the last 10 runs for every job,
    reducing the number of roundtrips to the DB, and then access them using the in-memory loader
    cache.
    """

    def __init__(self, instance: DagsterInstance, remote_repository: RemoteRepository):
        self._instance = instance
        self._repository = remote_repository
        self._data: dict[RepositoryDataType, dict[str, list[Any]]] = {}
        self._limits: dict[RepositoryDataType, int] = {}

    @property
    def repository(self) -> RemoteRepository:
        return self._repository

    def _get(self, data_type: RepositoryDataType, key: str, limit: int) -> Sequence[Any]:
        check.inst_param(data_type, "data_type", RepositoryDataType)
        check.str_param(key, "key")
        check.int_param(limit, "limit")
        if self._data.get(data_type) is None or limit > self._limits.get(data_type, 0):
            self._fetch(data_type, limit)
        return self._data[data_type].get(key, [])[:limit]

    def _fetch(self, data_type: RepositoryDataType, limit: int) -> None:
        check.inst_param(data_type, "data_type", RepositoryDataType)
        check.int_param(limit, "limit")

        fetched: dict[str, list[Any]] = defaultdict(list)

        if data_type == RepositoryDataType.SCHEDULE_STATES:
            schedule_states = self._instance.all_instigator_state(
                repository_origin_id=self._repository.get_remote_origin_id(),
                repository_selector_id=self._repository.selector_id,
                instigator_type=InstigatorType.SCHEDULE,
            )
            for state in schedule_states:
                fetched[state.name].append(state)

        elif data_type == RepositoryDataType.SENSOR_STATES:
            sensor_states = self._instance.all_instigator_state(
                repository_origin_id=self._repository.get_remote_origin_id(),
                repository_selector_id=self._repository.selector_id,
                instigator_type=InstigatorType.SENSOR,
            )
            for state in sensor_states:
                fetched[state.name].append(state)

        elif data_type == RepositoryDataType.SCHEDULE_TICKS:
            if self._instance.supports_batch_tick_queries:
                selector_ids = [
                    schedule.selector_id for schedule in self._repository.get_schedules()
                ]
                ticks_by_selector = self._instance.get_batch_ticks(selector_ids, limit=limit)
                for schedule in self._repository.get_schedules():
                    fetched[schedule.get_remote_origin_id()] = list(
                        ticks_by_selector.get(schedule.selector_id, [])
                    )
            else:
                for schedule in self._repository.get_schedules():
                    origin_id = schedule.get_remote_origin_id()
                    fetched[origin_id] = list(
                        self._instance.get_ticks(origin_id, schedule.selector_id, limit=limit)
                    )

        elif data_type == RepositoryDataType.SENSOR_TICKS:
            if self._instance.supports_batch_tick_queries:
                selector_ids = [schedule.selector_id for schedule in self._repository.get_sensors()]
                ticks_by_selector = self._instance.get_batch_ticks(selector_ids, limit=limit)
                for sensor in self._repository.get_sensors():
                    fetched[sensor.get_remote_origin_id()] = list(
                        ticks_by_selector.get(sensor.selector_id, [])
                    )
            else:
                for sensor in self._repository.get_sensors():
                    origin_id = sensor.get_remote_origin_id()
                    fetched[origin_id] = list(
                        self._instance.get_ticks(origin_id, sensor.selector_id, limit=limit)
                    )

        else:
            check.failed(f"Unknown data type for {self.__class__.__name__}: {data_type}")

        self._data[data_type] = fetched
        self._limits[data_type] = limit

    def get_schedule_state(self, schedule_name: str) -> InstigatorState | None:
        check.invariant(self._repository.has_schedule(schedule_name))
        states = self._get(RepositoryDataType.SCHEDULE_STATES, schedule_name, 1)
        return states[0] if states else None

    def get_sensor_state(self, sensor_name: str) -> InstigatorState | None:
        check.invariant(self._repository.has_sensor(sensor_name))
        states = self._get(RepositoryDataType.SENSOR_STATES, sensor_name, 1)
        return states[0] if states else None

    def get_sensor_ticks(self, origin_id: str, selector_id: str, limit: int) -> Sequence[Any]:
        check.invariant(
            any(selector_id == sensor.selector_id for sensor in self._repository.get_sensors())
        )
        return self._get(RepositoryDataType.SENSOR_TICKS, origin_id, limit)

    def get_schedule_ticks(self, origin_id: str, selector_id: str, limit: int) -> Sequence[Any]:
        check.invariant(
            any(
                selector_id == schedule.selector_id for schedule in self._repository.get_schedules()
            )
        )
        return self._get(RepositoryDataType.SCHEDULE_TICKS, origin_id, limit)



from collections.abc import Sequence
from typing import TYPE_CHECKING, Union, cast

import dagster._check as check
from dagster._core.definitions.selector import (
    JobSelector,
    PartitionsByAssetSelector,
    RepositorySelector,
)
from dagster._core.definitions.utils import check_valid_title
from dagster._core.errors import DagsterInvariantViolationError, DagsterUserCodeProcessError
from dagster._core.events import AssetKey
from dagster._core.execution.asset_backfill import create_asset_backfill_data_from_asset_partitions
from dagster._core.execution.backfill import (
    BULK_ACTION_TERMINAL_STATUSES,
    BulkActionStatus,
    PartitionBackfill,
)
from dagster._core.execution.job_backfill import submit_backfill_runs
from dagster._core.execution.plan.resume_retry import ReexecutionStrategy
from dagster._core.remote_representation.external_data import PartitionExecutionErrorSnap
from dagster._core.storage.tags import PARENT_BACKFILL_ID_TAG, ROOT_BACKFILL_ID_TAG
from dagster._core.utils import make_new_backfill_id
from dagster._core.workspace.permissions import Permissions
from dagster._time import datetime_from_timestamp, get_current_timestamp
from dagster._utils.caching_instance_queryer import CachingInstanceQueryer
from dagster_shared.error import DagsterError

from dagster_graphql.implementation.utils import (
    AssetBackfillPreviewParams,
    BackfillParams,
    assert_permission_for_asset_graph,
    assert_permission_for_backfill,
    assert_permission_for_job,
    assert_permission_for_location,
    assert_valid_asset_partition_backfill,
    assert_valid_job_partition_backfill,
)

BACKFILL_CHUNK_SIZE = 25


if TYPE_CHECKING:
    from dagster_graphql.schema.backfill import (
        GrapheneAssetPartitions,
        GrapheneCancelBackfillSuccess,
        GrapheneLaunchBackfillSuccess,
        GrapheneResumeBackfillSuccess,
    )
    from dagster_graphql.schema.errors import GraphenePartitionSetNotFoundError
    from dagster_graphql.schema.util import ResolveInfo


def get_asset_backfill_preview(
    graphene_info: "ResolveInfo", backfill_preview_params: AssetBackfillPreviewParams
) -> Sequence["GrapheneAssetPartitions"]:
    from dagster_graphql.schema.backfill import GrapheneAssetPartitions

    asset_graph = graphene_info.context.asset_graph

    check.invariant(backfill_preview_params.get("assetSelection") is not None)
    check.invariant(backfill_preview_params.get("partitionNames") is not None)

    asset_selection = [
        cast("AssetKey", AssetKey.from_graphql_input(asset_key))
        for asset_key in backfill_preview_params["assetSelection"]
    ]
    partition_names: list[str] = backfill_preview_params["partitionNames"]

    asset_backfill_data = create_asset_backfill_data_from_asset_partitions(
        asset_graph, asset_selection, partition_names, graphene_info.context.instance
    )

    asset_partitions = []

    for asset_key in asset_backfill_data.get_targeted_asset_keys_topological_order(asset_graph):
        if asset_graph.get(asset_key).partitions_def:
            partitions_subset = asset_backfill_data.target_subset.partitions_subsets_by_asset_key[
                asset_key
            ]
            asset_partitions.append(
                GrapheneAssetPartitions(asset_key=asset_key, partitions_subset=partitions_subset)
            )
        else:
            asset_partitions.append(
                GrapheneAssetPartitions(asset_key=asset_key, partitions_subset=None)
            )

    return asset_partitions


def create_and_launch_partition_backfill(
    graphene_info: "ResolveInfo",
    backfill_params: BackfillParams,
) -> Union["GrapheneLaunchBackfillSuccess", "GraphenePartitionSetNotFoundError"]:
    from dagster_graphql.schema.backfill import GrapheneLaunchBackfillSuccess
    from dagster_graphql.schema.errors import GraphenePartitionSetNotFoundError
    from dagster_graphql.schema.runs import parse_run_config_input

    backfill_id = make_new_backfill_id()
    backfill_timestamp = get_current_timestamp()
    backfill_datetime = datetime_from_timestamp(backfill_timestamp)
    dynamic_partitions_store = CachingInstanceQueryer(
        instance=graphene_info.context.instance,
        asset_graph=graphene_info.context.asset_graph,
        loading_context=graphene_info.context,
        evaluation_time=backfill_datetime,
    )

    asset_selection = (
        [
            cast("AssetKey", AssetKey.from_graphql_input(asset_key))
            for asset_key in backfill_params["assetSelection"]
        ]
        if backfill_params.get("assetSelection")
        else None
    )

    partitions_by_assets = backfill_params.get("partitionsByAssets")

    if (
        asset_selection or backfill_params.get("selector") or backfill_params.get("partitionNames")
    ) and partitions_by_assets:
        raise DagsterInvariantViolationError(
            "partitions_by_assets cannot be used together with asset_selection, selector, or"
            " partitionNames"
        )

    tags = {t["key"]: t["value"] for t in backfill_params.get("tags", [])}

    tags = {**tags, **graphene_info.context.get_viewer_tags()}

    title = check_valid_title(backfill_params.get("title"))
    parsed_run_config = (
        parse_run_config_input(backfill_params["runConfigData"], raise_on_error=False)
        if "runConfigData" in backfill_params
        else None
    )
    run_config = parsed_run_config if isinstance(parsed_run_config, dict) else None

    if backfill_params.get("selector") is not None:  # job backfill
        partition_set_selector = backfill_params["selector"]
        partition_set_name = partition_set_selector.get("partitionSetName")
        repository_selector = RepositorySelector.from_graphql_input(
            partition_set_selector.get("repositorySelector")
        )
        partitions_sets = graphene_info.context.get_partition_sets(repository_selector)
        matches = [
            partition_set
            for partition_set in partitions_sets
            if partition_set.name == partition_set_selector.get("partitionSetName")
        ]
        if len(matches) != 1:
            # if no matches, either throw a not found error or a more general invariant violation
            # but, to maintain the same behavior, we should check location permissions first, which
            # got deferred to support per-definition permissions
            assert_permission_for_location(
                graphene_info,
                Permissions.LAUNCH_PARTITION_BACKFILL,
                repository_selector.location_name,
            )
            if not matches:
                return GraphenePartitionSetNotFoundError(partition_set_name)

            raise DagsterInvariantViolationError(
                f"Partition set names must be unique: found {len(matches)} matches for {partition_set_name}"
            )
        remote_partition_set = next(iter(matches))
        assert_permission_for_job(
            graphene_info,
            Permissions.LAUNCH_PARTITION_BACKFILL,
            JobSelector(
                location_name=repository_selector.location_name,
                repository_name=repository_selector.repository_name,
                job_name=remote_partition_set.job_name,
            ),
        )

        if backfill_params.get("allPartitions"):
            result = graphene_info.context.get_partition_names(
                repository_selector=repository_selector,
                job_name=remote_partition_set.job_name,
                instance=graphene_info.context.instance,
                selected_asset_keys=None,
            )
            if isinstance(result, PartitionExecutionErrorSnap):
                raise DagsterUserCodeProcessError.from_error_info(result.error)
            partition_names = result.partition_names
        elif backfill_params.get("partitionNames"):
            partition_names = backfill_params["partitionNames"]
        else:
            raise DagsterError(
                'Backfill requested without specifying either "allPartitions" or "partitionNames" '
                "arguments"
            )

        backfill = PartitionBackfill(
            backfill_id=backfill_id,
            partition_set_origin=remote_partition_set.get_remote_origin(),
            status=BulkActionStatus.REQUESTED,
            partition_names=partition_names,
            from_failure=bool(backfill_params.get("fromFailure")),
            reexecution_steps=backfill_params.get("reexecutionSteps"),
            tags=tags,
            backfill_timestamp=backfill_timestamp,
            asset_selection=asset_selection,
            title=title,
            description=backfill_params.get("description"),
            run_config=run_config,
        )
        assert_valid_job_partition_backfill(
            graphene_info,
            backfill,
            remote_partition_set.get_partitions_definition(),
            dynamic_partitions_store,
            backfill_datetime,
        )

        if backfill_params.get("forceSynchronousSubmission"):
            # should only be used in a test situation
            to_submit = [name for name in partition_names]
            submitted_run_ids: list[str] = []

            while to_submit:
                chunk = to_submit[:BACKFILL_CHUNK_SIZE]
                to_submit = to_submit[BACKFILL_CHUNK_SIZE:]
                submitted_run_ids.extend(
                    run_id
                    for run_id in submit_backfill_runs(
                        graphene_info.context.instance,
                        create_workspace=lambda: graphene_info.context,
                        backfill_job=backfill,
                        partition_names_or_ranges=chunk,
                    )
                    if run_id is not None
                )
            return GrapheneLaunchBackfillSuccess(
                backfill_id=backfill_id, launched_run_ids=submitted_run_ids
            )
    elif asset_selection is not None:  # pure asset backfill
        if backfill_params.get("forceSynchronousSubmission"):
            raise DagsterError(
                "forceSynchronousSubmission is not supported for pure asset backfills"
            )

        if backfill_params.get("fromFailure"):
            raise DagsterError("fromFailure is not supported for pure asset backfills")

        asset_graph = graphene_info.context.asset_graph

        assert_permission_for_asset_graph(
            graphene_info, asset_graph, asset_selection, Permissions.LAUNCH_PARTITION_BACKFILL
        )

        backfill = PartitionBackfill.from_asset_partitions(
            backfill_id=backfill_id,
            asset_graph=asset_graph,
            backfill_timestamp=backfill_timestamp,
            tags=tags,
            asset_selection=asset_selection,
            partition_names=backfill_params.get("partitionNames"),
            dynamic_partitions_store=dynamic_partitions_store,
            all_partitions=backfill_params.get("allPartitions", False),
            title=title,
            description=backfill_params.get("description"),
            run_config=run_config,
        )
        assert_valid_asset_partition_backfill(
            graphene_info,
            backfill,
            backfill_datetime,
        )
    elif partitions_by_assets is not None:
        if backfill_params.get("forceSynchronousSubmission"):
            raise DagsterError(
                "forceSynchronousSubmission is not supported for pure asset backfills"
            )

        if backfill_params.get("fromFailure"):
            raise DagsterError("fromFailure is not supported for pure asset backfills")

        asset_graph = graphene_info.context.asset_graph

        partitions_by_assets = [
            PartitionsByAssetSelector.from_graphql_input(partitions_by_asset_selector)
            for partitions_by_asset_selector in partitions_by_assets
        ]

        selected_assets = list({selector.asset_key for selector in partitions_by_assets})

        assert_permission_for_asset_graph(
            graphene_info, asset_graph, selected_assets, Permissions.LAUNCH_PARTITION_BACKFILL
        )
        backfill = PartitionBackfill.from_partitions_by_assets(
            backfill_id=backfill_id,
            asset_graph=asset_graph,
            backfill_timestamp=backfill_timestamp,
            tags=tags,
            dynamic_partitions_store=dynamic_partitions_store,
            partitions_by_assets=partitions_by_assets,
            title=title,
            description=backfill_params.get("description"),
            run_config=run_config,
        )
        assert_valid_asset_partition_backfill(
            graphene_info,
            backfill,
            backfill_datetime,
        )

    else:
        raise DagsterError(
            "Backfill requested without specifying partition set selector or asset selection"
        )

    graphene_info.context.instance.add_backfill(backfill)
    return GrapheneLaunchBackfillSuccess(backfill_id=backfill_id)


def cancel_partition_backfill(
    graphene_info: "ResolveInfo", backfill_id: str
) -> "GrapheneCancelBackfillSuccess":
    from dagster_graphql.schema.backfill import GrapheneCancelBackfillSuccess

    backfill = graphene_info.context.instance.get_backfill(backfill_id)
    if not backfill:
        check.failed(f"No backfill found for id: {backfill_id}")

    assert_permission_for_backfill(graphene_info, Permissions.CANCEL_PARTITION_BACKFILL, backfill)

    graphene_info.context.instance.update_backfill(backfill.with_status(BulkActionStatus.CANCELING))

    return GrapheneCancelBackfillSuccess(backfill_id=backfill_id)


def resume_partition_backfill(
    graphene_info: "ResolveInfo", backfill_id: str
) -> "GrapheneResumeBackfillSuccess":
    from dagster_graphql.schema.backfill import GrapheneResumeBackfillSuccess

    backfill = graphene_info.context.instance.get_backfill(backfill_id)
    if not backfill:
        check.failed(f"No backfill found for id: {backfill_id}")

    assert_permission_for_backfill(graphene_info, Permissions.LAUNCH_PARTITION_BACKFILL, backfill)

    graphene_info.context.instance.update_backfill(backfill.with_status(BulkActionStatus.REQUESTED))
    return GrapheneResumeBackfillSuccess(backfill_id=backfill_id)


def retry_partition_backfill(
    graphene_info: "ResolveInfo", backfill_id: str, strategy: str
) -> "GrapheneLaunchBackfillSuccess":
    from dagster_graphql.schema.backfill import GrapheneLaunchBackfillSuccess

    backfill = graphene_info.context.instance.get_backfill(backfill_id)
    from_failure = ReexecutionStrategy(strategy) == ReexecutionStrategy.FROM_FAILURE
    if not backfill:
        check.failed(f"No backfill found for id: {backfill_id}")

    if backfill.status not in BULK_ACTION_TERMINAL_STATUSES:
        raise DagsterInvariantViolationError(
            f"Cannot re-execute backfill {backfill_id} because it is still in progress."
        )

    if backfill.is_asset_backfill:
        asset_backfill_data = backfill.get_asset_backfill_data(graphene_info.context.asset_graph)
        assets_to_request = asset_backfill_data.target_subset
        if from_failure:
            # determine the subset that should be retried by removing the successfully materialized subset from
            # the target subset. This ensures that if the backfill was canceled or marked failed that all
            # non-materialized asset partitions will be retried. asset_backfill_data.failed_and_downstream_asset
            # only contains asset partitions who's materialization runs failed and their downsteam assets, not
            # asset partitions that never got materialization runs.
            assets_to_request = assets_to_request - asset_backfill_data.materialized_subset
        if assets_to_request.num_partitions_and_non_partitioned_assets == 0:
            raise DagsterInvariantViolationError(
                "Cannot re-execute from failure an asset backfill that has no missing materializations."
            )

        assert_permission_for_backfill(
            graphene_info, Permissions.LAUNCH_PARTITION_BACKFILL, backfill
        )
        new_backfill = PartitionBackfill.from_asset_graph_subset(
            backfill_id=make_new_backfill_id(),
            asset_graph_subset=assets_to_request,
            dynamic_partitions_store=graphene_info.context.instance,
            tags={
                **backfill.tags,
                PARENT_BACKFILL_ID_TAG: backfill.backfill_id,
                ROOT_BACKFILL_ID_TAG: backfill.tags.get(ROOT_BACKFILL_ID_TAG, backfill.backfill_id),
            },
            backfill_timestamp=get_current_timestamp(),
            title=f"Re-execution of {backfill.title}" if backfill.title else None,
            description=backfill.description,
            run_config=backfill.run_config,
        )
    else:  # job backfill
        assert_permission_for_backfill(
            graphene_info, Permissions.LAUNCH_PARTITION_BACKFILL, backfill
        )

        new_backfill = PartitionBackfill(
            backfill_id=make_new_backfill_id(),
            partition_set_origin=backfill.partition_set_origin,
            status=BulkActionStatus.REQUESTED,
            partition_names=backfill.partition_names,
            from_failure=from_failure,
            reexecution_steps=backfill.reexecution_steps,
            tags={
                **backfill.tags,
                PARENT_BACKFILL_ID_TAG: backfill.backfill_id,
                ROOT_BACKFILL_ID_TAG: backfill.tags.get(ROOT_BACKFILL_ID_TAG, backfill.backfill_id),
            },
            backfill_timestamp=get_current_timestamp(),
            asset_selection=backfill.asset_selection,
            title=f"Re-execution of {backfill.title}" if backfill.title else None,
            description=backfill.description,
        )

    graphene_info.context.instance.add_backfill(new_backfill)
    return GrapheneLaunchBackfillSuccess(backfill_id=new_backfill.backfill_id)



from collections.abc import Sequence

import graphene
from dagster._core.definitions.auto_materialize_rule_evaluation import AutoMaterializeDecisionType
from dagster._core.definitions.declarative_automation.legacy.rule_condition import RuleCondition
from dagster._core.definitions.declarative_automation.serialized_objects import (
    AssetSubsetWithMetadata,
    AutomationConditionEvaluation,
)
from dagster._core.definitions.metadata import DagsterAssetMetadataValue
from dagster._core.scheduler.instigation import AutoMaterializeAssetEvaluationRecord

from dagster_graphql.schema.auto_materialize_policy import GrapheneAutoMaterializeRule
from dagster_graphql.schema.entity_key import GrapheneAssetKey
from dagster_graphql.schema.errors import GrapheneError
from dagster_graphql.schema.partition_keys import (
    GraphenePartitionKeys,
    GraphenePartitionKeysOrError,
)
from dagster_graphql.schema.util import non_null_list

GrapheneAutoMaterializeDecisionType = graphene.Enum.from_enum(AutoMaterializeDecisionType)


class GrapheneTextRuleEvaluationData(graphene.ObjectType):
    text = graphene.String()

    class Meta:
        name = "TextRuleEvaluationData"


class GrapheneParentMaterializedRuleEvaluationData(graphene.ObjectType):
    updatedAssetKeys = graphene.List(graphene.NonNull(GrapheneAssetKey))
    willUpdateAssetKeys = graphene.List(graphene.NonNull(GrapheneAssetKey))

    class Meta:
        name = "ParentMaterializedRuleEvaluationData"


class GrapheneWaitingOnKeysRuleEvaluationData(graphene.ObjectType):
    waitingOnAssetKeys = graphene.List(graphene.NonNull(GrapheneAssetKey))

    class Meta:
        name = "WaitingOnKeysRuleEvaluationData"


class GrapheneAutoMaterializeRuleEvaluationData(graphene.Union):
    class Meta:
        types = (
            GrapheneTextRuleEvaluationData,
            GrapheneParentMaterializedRuleEvaluationData,
            GrapheneWaitingOnKeysRuleEvaluationData,
        )
        name = "AutoMaterializeRuleEvaluationData"


class GrapheneAutoMaterializeRuleEvaluation(graphene.ObjectType):
    partitionKeysOrError = graphene.Field(GraphenePartitionKeysOrError)
    evaluationData = graphene.Field(GrapheneAutoMaterializeRuleEvaluationData)

    class Meta:
        name = "AutoMaterializeRuleEvaluation"


class GrapheneAutoMaterializeRuleWithRuleEvaluations(graphene.ObjectType):
    rule = graphene.NonNull(GrapheneAutoMaterializeRule)
    ruleEvaluations = non_null_list(GrapheneAutoMaterializeRuleEvaluation)

    class Meta:
        name = "AutoMaterializeRuleWithRuleEvaluations"

    def __init__(
        self,
        rule: GrapheneAutoMaterializeRule,
        ruleEvaluations,
    ):
        super().__init__(
            rule=rule,
            ruleEvaluations=ruleEvaluations,
        )


def create_graphene_auto_materialize_rule_evaluation(
    asset_subset_with_metadata: AssetSubsetWithMetadata,
) -> GrapheneAutoMaterializeRuleEvaluation | None:
    if not asset_subset_with_metadata.subset.is_partitioned:
        partition_keys_or_error = None
    else:
        partition_keys_or_error = GraphenePartitionKeys(
            partitionKeys=asset_subset_with_metadata.subset.subset_value.get_partition_keys()
        )

    metadata = asset_subset_with_metadata.metadata
    if "text" in metadata.keys() and isinstance(metadata["text"], str):
        rule_evaluation_data = GrapheneTextRuleEvaluationData(text=metadata["text"])
    elif any(key.startswith("updated_parent") for key in metadata.keys()):
        updatedAssetKeys = [
            value.asset_key
            for key, value in sorted(metadata.items())
            if key.startswith("updated_parent") and isinstance(value, DagsterAssetMetadataValue)
        ]
        willUpdateAssetKeys = [
            value.asset_key
            for key, value in sorted(metadata.items())
            if key.startswith("will_update_parent") and isinstance(value, DagsterAssetMetadataValue)
        ]
        rule_evaluation_data = GrapheneParentMaterializedRuleEvaluationData(
            updatedAssetKeys=updatedAssetKeys, willUpdateAssetKeys=willUpdateAssetKeys
        )
    elif any(key.startswith("waiting_on_ancestor") for key in metadata.keys()):
        waitingOnAssetKeys = [
            value.asset_key
            for key, value in sorted(metadata.items())
            if key.startswith("waiting_on_ancestor")
            and isinstance(value, DagsterAssetMetadataValue)
        ]
        rule_evaluation_data = GrapheneWaitingOnKeysRuleEvaluationData(
            waitingOnAssetKeys=waitingOnAssetKeys
        )
    else:
        rule_evaluation_data = None

    return GrapheneAutoMaterializeRuleEvaluation(
        partitionKeysOrError=partition_keys_or_error, evaluationData=rule_evaluation_data
    )


def _create_rules_with_rule_evaluations_for_decision_type(
    evaluation: AutomationConditionEvaluation, decision_type: AutoMaterializeDecisionType
) -> tuple[
    Sequence[GrapheneAutoMaterializeRule], Sequence[GrapheneAutoMaterializeRuleWithRuleEvaluations]
]:
    rules = []
    rules_with_rule_evaluations = []
    leaf_evaluations = evaluation.child_evaluations
    for le in leaf_evaluations:
        snapshot = le.condition_snapshot
        if snapshot.class_name != RuleCondition.__name__:
            continue
        rule = GrapheneAutoMaterializeRule(snapshot.description, decision_type)
        rules.append(rule)
        if le.subsets_with_metadata:
            rules_with_rule_evaluations.append(
                GrapheneAutoMaterializeRuleWithRuleEvaluations(
                    rule=rule,
                    ruleEvaluations=[
                        create_graphene_auto_materialize_rule_evaluation(sswm)
                        for sswm in le.subsets_with_metadata
                    ],
                )
            )
        elif le.true_subset.size > 0:
            rules_with_rule_evaluations.append(
                GrapheneAutoMaterializeRuleWithRuleEvaluations(
                    rule=rule,
                    ruleEvaluations=[
                        GrapheneAutoMaterializeRuleEvaluation(
                            partitionKeysOrError=GraphenePartitionKeys(
                                partitionKeys=le.true_subset.subset_value.get_partition_keys()
                            )
                            if le.true_subset.is_partitioned
                            else None,
                            evaluationData=None,
                        )
                    ],
                )
            )
    return rules, rules_with_rule_evaluations


def create_graphene_auto_materialize_rules_with_rule_evaluations(
    evaluation: AutomationConditionEvaluation,
) -> tuple[
    Sequence[GrapheneAutoMaterializeRule], Sequence[GrapheneAutoMaterializeRuleWithRuleEvaluations]
]:
    rules, rules_with_rule_evaluations = [], []

    if len(evaluation.child_evaluations) > 0:
        materialize_evaluation = evaluation.child_evaluations[0]
        rs, rwres = _create_rules_with_rule_evaluations_for_decision_type(
            materialize_evaluation, AutoMaterializeDecisionType.MATERIALIZE
        )
        rules.extend(rs)
        rules_with_rule_evaluations.extend(rwres)

    if (
        len(evaluation.child_evaluations) > 1
        and len(evaluation.child_evaluations[1].child_evaluations) == 1
    ):
        skip_evaluation = evaluation.child_evaluations[1].child_evaluations[0]
        rs, rwres = _create_rules_with_rule_evaluations_for_decision_type(
            skip_evaluation, AutoMaterializeDecisionType.SKIP
        )
        rules.extend(rs)
        rules_with_rule_evaluations.extend(rwres)

    if (
        len(evaluation.child_evaluations) > 2
        and len(evaluation.child_evaluations[2].child_evaluations) == 1
    ):
        discard_evaluation = evaluation.child_evaluations[2]
        rs, rwres = _create_rules_with_rule_evaluations_for_decision_type(
            discard_evaluation, AutoMaterializeDecisionType.DISCARD
        )
        rules.extend(rs)
        rules_with_rule_evaluations.extend(rwres)

    return rules, rules_with_rule_evaluations


class GrapheneAutoMaterializeAssetEvaluationRecord(graphene.ObjectType):
    id = graphene.NonNull(graphene.ID)
    evaluationId = graphene.NonNull(graphene.ID)
    numRequested = graphene.NonNull(graphene.Int)
    numSkipped = graphene.NonNull(graphene.Int)
    numDiscarded = graphene.NonNull(graphene.Int)
    rulesWithRuleEvaluations = non_null_list(GrapheneAutoMaterializeRuleWithRuleEvaluations)
    timestamp = graphene.NonNull(graphene.Float)
    runIds = non_null_list(graphene.String)
    rules = graphene.Field(graphene.List(graphene.NonNull(GrapheneAutoMaterializeRule)))
    assetKey = graphene.NonNull(GrapheneAssetKey)

    class Meta:
        name = "AutoMaterializeAssetEvaluationRecord"

    def __init__(self, record: AutoMaterializeAssetEvaluationRecord):
        evaluation_with_run_ids = record.get_evaluation_with_run_ids()
        evaluation = evaluation_with_run_ids.evaluation
        (
            rules,
            rules_with_rule_evaluations,
        ) = create_graphene_auto_materialize_rules_with_rule_evaluations(evaluation)
        super().__init__(
            id=record.id,
            evaluationId=record.evaluation_id,
            numRequested=evaluation_with_run_ids.evaluation.true_subset.size,
            numSkipped=0,
            numDiscarded=0,
            rulesWithRuleEvaluations=rules_with_rule_evaluations,
            timestamp=record.timestamp,
            runIds=evaluation_with_run_ids.run_ids,
            rules=sorted(rules, key=lambda rule: rule.className),
            assetKey=GrapheneAssetKey(path=record.key.path),
        )


class GrapheneAutoMaterializeAssetEvaluationRecords(graphene.ObjectType):
    records = non_null_list(GrapheneAutoMaterializeAssetEvaluationRecord)

    class Meta:
        name = "AutoMaterializeAssetEvaluationRecords"


class GrapheneAutoMaterializeAssetEvaluationNeedsMigrationError(graphene.ObjectType):
    message = graphene.NonNull(graphene.String)

    class Meta:
        interfaces = (GrapheneError,)
        name = "AutoMaterializeAssetEvaluationNeedsMigrationError"


class GrapheneAutoMaterializeAssetEvaluationRecordsOrError(graphene.Union):
    class Meta:
        types = (
            GrapheneAutoMaterializeAssetEvaluationRecords,
            GrapheneAutoMaterializeAssetEvaluationNeedsMigrationError,
        )
        name = "AutoMaterializeAssetEvaluationRecordsOrError"



from typing import Any
from unittest.mock import patch

from dagster._core.test_utils import wait_for_runs_to_finish
from dagster._core.workspace.context import WorkspaceRequestContext
from dagster_graphql.client.query import (
    LAUNCH_MULTIPLE_RUNS_MUTATION,
    LAUNCH_PIPELINE_EXECUTION_MUTATION,
)
from dagster_graphql.test.utils import GqlResult, execute_dagster_graphql, infer_job_selector

from dagster_graphql_tests.graphql.graphql_context_test_suite import (
    GraphQLContextVariant,
    make_graphql_context_test_suite,
)

RUN_QUERY = """
query RunQuery($runId: ID!) {
  pipelineRunOrError(runId: $runId) {
    __typename
    ... on Run {
      status
      stats {
        ... on RunStatsSnapshot {
          stepsSucceeded
        }
      }
      startTime
      endTime
    }
  }
}
"""


BaseTestSuite: Any = make_graphql_context_test_suite(
    context_variants=GraphQLContextVariant.all_executing_variants()
)
LaunchFailTestSuite: Any = make_graphql_context_test_suite(
    context_variants=GraphQLContextVariant.all_non_launchable_variants()
)
ReadOnlyTestSuite: Any = make_graphql_context_test_suite(
    context_variants=GraphQLContextVariant.all_readonly_variants()
)


class TestBasicLaunch(BaseTestSuite):
    def test_run_launcher(self, graphql_context: WorkspaceRequestContext):
        selector = infer_job_selector(graphql_context, "no_config_job")
        result = execute_dagster_graphql(
            context=graphql_context,
            query=LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={"executionParams": {"selector": selector, "mode": "default"}},
        )

        assert result.data["launchPipelineExecution"]["__typename"] == "LaunchRunSuccess"
        assert result.data["launchPipelineExecution"]["run"]["status"] == "STARTING"

        run_id = result.data["launchPipelineExecution"]["run"]["runId"]

        wait_for_runs_to_finish(graphql_context.instance)

        result = execute_dagster_graphql(
            context=graphql_context, query=RUN_QUERY, variables={"runId": run_id}
        )
        assert result.data["pipelineRunOrError"]["__typename"] == "Run"
        assert result.data["pipelineRunOrError"]["status"] == "SUCCESS"

    def test_run_launcher_subset(self, graphql_context: WorkspaceRequestContext):
        selector = infer_job_selector(graphql_context, "more_complicated_config", ["noop_op"])
        result = execute_dagster_graphql(
            context=graphql_context,
            query=LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={
                "executionParams": {
                    "selector": selector,
                    "mode": "default",
                }
            },
        )

        assert result.data["launchPipelineExecution"]["__typename"] == "LaunchRunSuccess"
        assert result.data["launchPipelineExecution"]["run"]["status"] == "STARTING"

        run_id = result.data["launchPipelineExecution"]["run"]["runId"]

        wait_for_runs_to_finish(graphql_context.instance)

        result = execute_dagster_graphql(
            context=graphql_context, query=RUN_QUERY, variables={"runId": run_id}
        )
        assert result.data["pipelineRunOrError"]["__typename"] == "Run"
        assert result.data["pipelineRunOrError"]["status"] == "SUCCESS"
        assert result.data["pipelineRunOrError"]["stats"]["stepsSucceeded"] == 1

    def test_run_launcher_unauthorized(self, graphql_context: WorkspaceRequestContext):
        selector = infer_job_selector(graphql_context, "no_config_job")

        with (
            patch.object(graphql_context, "has_permission_for_selector", return_value=False),
            patch.object(graphql_context, "was_permission_checked", return_value=True),
        ):
            result = execute_dagster_graphql(
                context=graphql_context,
                query=LAUNCH_PIPELINE_EXECUTION_MUTATION,
                variables={"executionParams": {"selector": selector, "mode": "default"}},
            )
            assert result.data["launchPipelineExecution"]["__typename"] == "UnauthorizedError"


class TestMultipleLaunch(BaseTestSuite):
    def test_multiple_run_launcher_same_job(self, graphql_context: WorkspaceRequestContext):
        selector = infer_job_selector(graphql_context, "no_config_job")

        # test with multiple of the same job
        executionParamsList = [
            {"selector": selector, "mode": "default"},
            {"selector": selector, "mode": "default"},
            {"selector": selector, "mode": "default"},
        ]

        result = execute_dagster_graphql(
            context=graphql_context,
            query=LAUNCH_MULTIPLE_RUNS_MUTATION,
            variables={"executionParamsList": executionParamsList},
        )

        assert "launchMultipleRuns" in result.data
        launches = result.data["launchMultipleRuns"]

        assert launches["__typename"] == "LaunchMultipleRunsResult"
        assert "launchMultipleRunsResult" in launches
        results = launches["launchMultipleRunsResult"]

        for result in results:
            assert result["__typename"] == "LaunchRunSuccess"

    def test_multiple_run_launcher_multiple_jobs(self, graphql_context: WorkspaceRequestContext):
        selectors = [
            infer_job_selector(graphql_context, "no_config_job"),
            infer_job_selector(graphql_context, "more_complicated_config", ["noop_op"]),
        ]

        # test with multiple of the same job
        executionParamsList = [
            {"selector": selectors[0], "mode": "default"},
            {"selector": selectors[1], "mode": "default"},
            {"selector": selectors[0], "mode": "default"},
            {"selector": selectors[1], "mode": "default"},
        ]

        result = execute_dagster_graphql(
            context=graphql_context,
            query=LAUNCH_MULTIPLE_RUNS_MUTATION,
            variables={"executionParamsList": executionParamsList},
        )

        assert "launchMultipleRuns" in result.data
        launches = result.data["launchMultipleRuns"]

        assert launches["__typename"] == "LaunchMultipleRunsResult"
        assert "launchMultipleRunsResult" in launches
        results = launches["launchMultipleRunsResult"]

        for result in results:
            assert result["__typename"] == "LaunchRunSuccess"

    def test_multiple_launch_failure_unauthorized(self, graphql_context: WorkspaceRequestContext):
        executionParamsList = [
            {"selector": infer_job_selector(graphql_context, "no_config_job"), "mode": "default"},
            {"selector": infer_job_selector(graphql_context, "no_config_job"), "mode": "default"},
        ]

        # mock no permissions
        with patch.object(graphql_context, "has_permission_for_selector", return_value=False):
            result = execute_dagster_graphql(
                context=graphql_context,
                query=LAUNCH_MULTIPLE_RUNS_MUTATION,
                variables={"executionParamsList": executionParamsList},
            )
            assert "launchMultipleRuns" in result.data
            result_data = result.data["launchMultipleRuns"]

            assert result_data["__typename"] == "LaunchMultipleRunsResult"
            assert "launchMultipleRunsResult" in result_data

            results = result_data["launchMultipleRunsResult"]

            for result in results:
                assert result["__typename"] == "UnauthorizedError"


class TestFailedLaunch(LaunchFailTestSuite):
    def test_launch_failure(self, graphql_context: WorkspaceRequestContext):
        selector = infer_job_selector(graphql_context, "no_config_job")
        result = execute_dagster_graphql(
            context=graphql_context,
            query=LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={"executionParams": {"selector": selector, "mode": "default"}},
        )

        assert result.data["launchPipelineExecution"]["__typename"] != "LaunchRunSuccess"

        # fetch the most recent run, which should be this one that just failed to launch
        run = graphql_context.instance.get_runs(limit=1)[0]

        result = execute_dagster_graphql(
            context=graphql_context, query=RUN_QUERY, variables={"runId": run.run_id}
        )

        assert result.data["pipelineRunOrError"]["__typename"] == "Run"
        assert result.data["pipelineRunOrError"]["status"] == "FAILURE"
        assert result.data["pipelineRunOrError"]["startTime"]
        assert result.data["pipelineRunOrError"]["endTime"]


class TestFailedMultipleLaunch(LaunchFailTestSuite):
    def test_multiple_launch_failure(self, graphql_context: WorkspaceRequestContext):
        executionParamsList = [
            {"selector": infer_job_selector(graphql_context, "no_config_job"), "mode": "default"},
            {"selector": infer_job_selector(graphql_context, "no_config_job"), "mode": "default"},
        ]

        result = execute_dagster_graphql(
            context=graphql_context,
            query=LAUNCH_MULTIPLE_RUNS_MUTATION,
            variables={"executionParamsList": executionParamsList},
        )

        assert "launchMultipleRuns" in result.data
        result_data = result.data["launchMultipleRuns"]

        assert result_data["__typename"] == "LaunchMultipleRunsResult"
        results = result_data["launchMultipleRunsResult"]

        assert len(results) == 2

        for run_result in results:
            assert run_result["__typename"] == "PythonError"
            assert run_result["message"].startswith(
                "NotImplementedError: The entire purpose of this is to throw on launch"
            )
            assert run_result["className"] == "NotImplementedError"


class TestFailedMultipleLaunchReadOnly(ReadOnlyTestSuite):
    def test_multiple_launch_failure_readonly(self, graphql_context: WorkspaceRequestContext):
        executionParamsList = [
            {"selector": infer_job_selector(graphql_context, "no_config_job"), "mode": "default"},
            {"selector": infer_job_selector(graphql_context, "no_config_job"), "mode": "default"},
        ]

        result = execute_dagster_graphql(
            context=graphql_context,
            query=LAUNCH_MULTIPLE_RUNS_MUTATION,
            variables={"executionParamsList": executionParamsList},
        )
        assert "launchMultipleRuns" in result.data
        result_data = result.data["launchMultipleRuns"]

        assert result_data["__typename"] == "LaunchMultipleRunsResult"
        assert "launchMultipleRunsResult" in result_data

        results = result_data["launchMultipleRunsResult"]

        for result in results:
            assert result["__typename"] == "UnauthorizedError"


class TestSuccessAndFailureMultipleLaunch(BaseTestSuite):
    def test_launch_multiple_runs_success_and_failure(
        self, graphql_context: WorkspaceRequestContext
    ):
        launchSuccessExecutionParams = [
            {
                "selector": {
                    "repositoryLocationName": "test_location",
                    "repositoryName": "test_repo",
                    "pipelineName": "no_config_job",
                    "solidSelection": None,
                    "assetSelection": None,
                    "assetCheckSelection": None,
                },
                "mode": "default",
            },
            {
                "selector": {
                    "repositoryLocationName": "test_location",
                    "repositoryName": "test_repo",
                    "pipelineName": "no_config_job",
                    "solidSelection": None,
                    "assetSelection": None,
                    "assetCheckSelection": None,
                },
                "mode": "default",
            },
        ]

        pipelineNotFoundExecutionParams = [
            {
                "selector": {
                    "repositoryLocationName": "test_location",
                    "repositoryName": "test_dict_repo",
                    "pipelineName": "no_config_job",
                    "solidSelection": None,
                    "assetSelection": None,
                    "assetCheckSelection": None,
                },
                "mode": "default",
            },
            {
                "selector": {
                    "repositoryLocationName": "test_location",
                    "repositoryName": "test_dict_repo",
                    "pipelineName": "no_config_job",
                    "solidSelection": None,
                    "assetSelection": None,
                    "assetCheckSelection": None,
                },
                "mode": "default",
            },
        ]

        executionParamsList = [executionParams for executionParams in launchSuccessExecutionParams]
        executionParamsList.extend(
            [executionParams for executionParams in pipelineNotFoundExecutionParams]
        )

        result: GqlResult = execute_dagster_graphql(
            context=graphql_context,
            query=LAUNCH_MULTIPLE_RUNS_MUTATION,
            variables={"executionParamsList": executionParamsList},
        )

        assert "launchMultipleRuns" in result.data
        result_data = result.data["launchMultipleRuns"]

        assert result_data["__typename"] == "LaunchMultipleRunsResult"
        results = result_data["launchMultipleRunsResult"]

        assert len(results) == 4

        assert results[0]["__typename"] == "LaunchRunSuccess"
        assert results[1]["__typename"] == "LaunchRunSuccess"
        assert results[2]["__typename"] == "PipelineNotFoundError"
        assert results[3]["__typename"] == "PipelineNotFoundError"



from collections import namedtuple
from unittest.mock import patch

import pytest
from dagster_graphql import DagsterGraphQLClient

MockClient = namedtuple("MockClient", ("python_client", "mock_gql_client"))


@pytest.fixture(scope="function", name="mock_client")
def create_mock_client():
    with patch("dagster_graphql.client.client.Client") as _:
        client = DagsterGraphQLClient("localhost")
        yield MockClient(
            python_client=client,
            mock_gql_client=client._client,  # noqa: SLF001
        )


python_client_test_suite = pytest.mark.python_client_test_suite



from dagster._core.workspace.context import WorkspaceRequestContext
from dagster_graphql.test.utils import execute_dagster_graphql

from dagster_graphql_tests.graphql.graphql_context_test_suite import (
    ExecutingGraphQLContextTestMatrix,
)

GET_ASSET_HEALTH = """
query GetAssetHealth($assetKey: AssetKeyInput!) {
    assetsOrError(assetKeys: [$assetKey]) {
        ... on AssetConnection {
            nodes {
                assetHealth {
                    assetHealth
                    materializationStatus
                    assetChecksStatus
                    freshnessStatus
                }
            }
        }
    }
}
"""


# There is a separate implementation for plus graphql tests.
class TestAssetHealth(ExecutingGraphQLContextTestMatrix):
    def test_asset_health_status(self, graphql_context: WorkspaceRequestContext):
        instance = graphql_context.instance
        assert not instance.dagster_asset_health_queries_supported()
        res = execute_dagster_graphql(
            graphql_context, GET_ASSET_HEALTH, variables={"assetKey": {"path": ["asset_1"]}}
        )
        assert res.data == {"assetsOrError": {"nodes": [{"assetHealth": None}]}}



import json

from dagster._core.test_utils import poll_for_finished_run
from dagster._core.workspace.context import WorkspaceRequestContext
from dagster._utils import file_relative_path
from dagster_graphql.client.query import LAUNCH_PIPELINE_EXECUTION_MUTATION
from dagster_graphql.test.utils import execute_dagster_graphql, infer_job_selector

from dagster_graphql_tests.graphql.graphql_context_test_suite import (
    ExecutingGraphQLContextTestMatrix,
)
from dagster_graphql_tests.graphql.utils import sync_execute_get_events

GET_EXPECTATIONS_FROM_STEP_STATS = """
query MaterializationsFromStepStatsQuery($runId: ID!) {
  runOrError(runId: $runId) {
    ... on PythonError {
      className
      message
      stack
    }
    ... on Run {
      stepStats {
        expectationResults {
            success
            label
            description
            metadataEntries {
                ... on TextMetadataEntry {
                    text
                }
                ... on JsonMetadataEntry {
                    jsonString
                }
            }
        }
      }
    }
  }
}
"""


def _create_run(
    graphql_context: WorkspaceRequestContext,
    pipeline_name: str,
) -> str:
    selector = infer_job_selector(
        graphql_context,
        pipeline_name,
    )
    result = execute_dagster_graphql(
        graphql_context,
        LAUNCH_PIPELINE_EXECUTION_MUTATION,
        variables={
            "executionParams": {
                "selector": selector,
            }
        },
    )
    assert result.data["launchPipelineExecution"]["__typename"] == "LaunchRunSuccess"
    run_id = result.data["launchPipelineExecution"]["run"]["runId"]
    poll_for_finished_run(graphql_context.instance, run_id)
    return run_id


def get_expectation_results(logs, op_name: str):
    def _f():
        for log in logs:
            if log["__typename"] == "StepExpectationResultEvent" and log["stepKey"] == f"{op_name}":
                yield log

    return list(_f())


def get_expectation_result(logs, op_name: str):
    expt_results = get_expectation_results(logs, op_name)
    if len(expt_results) != 1:
        raise Exception("Only expected one expectation result")
    return expt_results[0]


class TestExpectations(ExecutingGraphQLContextTestMatrix):
    def test_basic_expectations_within_compute_step_events(
        self, graphql_context: WorkspaceRequestContext, snapshot
    ):
        selector = infer_job_selector(graphql_context, "job_with_expectations")
        logs = sync_execute_get_events(
            context=graphql_context,
            variables={"executionParams": {"selector": selector}},
        )

        emit_failed_expectation_event = get_expectation_result(logs, "emit_failed_expectation")
        assert emit_failed_expectation_event["expectationResult"]["success"] is False
        assert emit_failed_expectation_event["expectationResult"]["description"] == "Failure"
        failed_result_metadata = json.loads(
            emit_failed_expectation_event["expectationResult"]["metadataEntries"][0]["jsonString"]
        )
        assert emit_failed_expectation_event["expectationResult"]["label"] == "always_false"

        assert failed_result_metadata == {"reason": "Relentless pessimism."}

        emit_successful_expectation_event = get_expectation_result(
            logs, "emit_successful_expectation"
        )

        assert emit_successful_expectation_event["expectationResult"]["success"] is True
        assert emit_successful_expectation_event["expectationResult"]["description"] == "Successful"
        assert emit_successful_expectation_event["expectationResult"]["label"] == "always_true"
        successful_result_metadata = json.loads(
            emit_successful_expectation_event["expectationResult"]["metadataEntries"][0][
                "jsonString"
            ]
        )

        assert successful_result_metadata == {"reason": "Just because."}

        emit_no_metadata = get_expectation_result(logs, "emit_successful_expectation_no_metadata")
        assert not emit_no_metadata["expectationResult"]["metadataEntries"]

        snapshot.assert_match(get_expectation_results(logs, "emit_failed_expectation"))
        snapshot.assert_match(get_expectation_results(logs, "emit_successful_expectation"))
        snapshot.assert_match(
            get_expectation_results(logs, "emit_successful_expectation_no_metadata")
        )

    def test_get_expectation_results_from_step_stats(
        self, graphql_context: WorkspaceRequestContext
    ):
        run_id = _create_run(graphql_context, "job_with_expectations")
        result = execute_dagster_graphql(
            graphql_context, GET_EXPECTATIONS_FROM_STEP_STATS, {"runId": run_id}
        )
        assert result.data
        assert any(
            len(step["expectationResults"]) > 0
            and step["expectationResults"][0]
            == {
                "success": False,
                "label": "always_false",
                "description": "Failure",
                "metadataEntries": [
                    {"jsonString": json.dumps({"reason": "Relentless pessimism."})}
                ],
            }
            for step in result.data["runOrError"]["stepStats"]
        )

    def test_basic_input_output_expectations(
        self, graphql_context: WorkspaceRequestContext, snapshot
    ):
        selector = infer_job_selector(graphql_context, "csv_hello_world_with_expectations")
        logs = sync_execute_get_events(
            context=graphql_context,
            variables={
                "executionParams": {
                    "selector": selector,
                    "runConfigData": {
                        "ops": {
                            "sum_op": {
                                "inputs": {"num": file_relative_path(__file__, "../data/num.csv")}
                            }
                        }
                    },
                }
            },
        )

        expectation_results = get_expectation_results(logs, "df_expectations_op")
        assert len(expectation_results) == 2

        snapshot.assert_match(expectation_results)



import asyncio
import os
import sys
from collections.abc import AsyncIterator, Mapping, Sequence
from typing import TYPE_CHECKING, Any, List, Optional, Tuple, Union  # noqa: F401, UP035

# re-exports
import dagster._check as check
from dagster._annotations import deprecated
from dagster._core.definitions.asset_checks.asset_check_evaluation import AssetCheckEvaluation
from dagster._core.definitions.events import AssetKey, AssetPartitionWipeRange
from dagster._core.definitions.partitions.context import partition_loading_context
from dagster._core.events import (
    AssetMaterialization,
    AssetObservation,
    DagsterEventType,
    EngineEventData,
)
from dagster._core.events.log import EventLogEntry
from dagster._core.instance import DagsterInstance
from dagster._core.storage.dagster_run import CANCELABLE_RUN_STATUSES
from dagster._core.workspace.permissions import Permissions
from dagster._utils.error import serializable_error_info_from_exc_info
from starlette.concurrency import (
    run_in_threadpool,  # can provide this indirectly if we dont want starlette dep in dagster-graphql
)

if TYPE_CHECKING:
    from dagster._core.definitions.asset_checks.asset_check_spec import AssetCheckSeverity
    from dagster._core.definitions.metadata import RawMetadataValue

    from dagster_graphql.schema.errors import (
        GrapheneAssetNotFoundError,
        GrapheneUnauthorizedError,
        GrapheneUnsupportedOperationError,
    )
    from dagster_graphql.schema.roots.mutation import (
        GrapheneReportAssetCheckEvaluationSuccess,
        GrapheneTerminateRunPolicy,
    )


from dagster_graphql.implementation.execution.backfill import (
    cancel_partition_backfill as cancel_partition_backfill,
    create_and_launch_partition_backfill as create_and_launch_partition_backfill,
    resume_partition_backfill as resume_partition_backfill,
    retry_partition_backfill as retry_partition_backfill,
)
from dagster_graphql.implementation.utils import (
    assert_permission,
    assert_permission_for_run,
    has_permission_for_run,
)

if TYPE_CHECKING:
    from dagster._core.storage.compute_log_manager import CapturedLogData

    from dagster_graphql.schema.errors import GrapheneRunNotFoundError
    from dagster_graphql.schema.logs.compute_logs import GrapheneCapturedLogs
    from dagster_graphql.schema.pipelines.subscription import (
        GraphenePipelineRunLogsSubscriptionFailure,
        GraphenePipelineRunLogsSubscriptionSuccess,
    )
    from dagster_graphql.schema.roots.mutation import (
        GrapheneAssetWipeSuccess,
        GrapheneDeletePipelineRunSuccess,
        GrapheneReportRunlessAssetEventsSuccess,
        GrapheneTerminateRunFailure,
        GrapheneTerminateRunsResult,
        GrapheneTerminateRunSuccess,
    )
    from dagster_graphql.schema.util import ResolveInfo


def _force_mark_as_canceled(
    instance: DagsterInstance, run_id: str
) -> "GrapheneTerminateRunSuccess":
    from dagster_graphql.schema.pipelines.pipeline import GrapheneRun
    from dagster_graphql.schema.roots.mutation import GrapheneTerminateRunSuccess

    reloaded_record = check.not_none(instance.get_run_record_by_id(run_id))

    if not reloaded_record.dagster_run.is_finished:
        message = (
            "This pipeline was forcibly marked as canceled from outside the execution context. The "
            "computational resources created by the run may not have been fully cleaned up."
        )
        instance.report_run_canceled(reloaded_record.dagster_run, message=message)
        reloaded_record = check.not_none(instance.get_run_record_by_id(run_id))

    return GrapheneTerminateRunSuccess(GrapheneRun(reloaded_record))


def terminate_pipeline_execution(
    graphene_info: "ResolveInfo",
    run_id: str,
    terminate_policy: "GrapheneTerminateRunPolicy",
) -> Union[
    "GrapheneTerminateRunSuccess", "GrapheneTerminateRunFailure", "GrapheneUnauthorizedError"
]:
    from dagster_graphql.schema.errors import GrapheneRunNotFoundError, GrapheneUnauthorizedError
    from dagster_graphql.schema.pipelines.pipeline import GrapheneRun
    from dagster_graphql.schema.roots.mutation import (
        GrapheneTerminateRunFailure,
        GrapheneTerminateRunPolicy,
        GrapheneTerminateRunSuccess,
    )

    check.str_param(run_id, "run_id")

    instance = graphene_info.context.instance

    record = instance.get_run_record_by_id(run_id)

    force_mark_as_canceled = (
        terminate_policy == GrapheneTerminateRunPolicy.MARK_AS_CANCELED_IMMEDIATELY
    )

    if not record:
        if not graphene_info.context.has_permission(Permissions.TERMINATE_PIPELINE_EXECUTION):
            return GrapheneUnauthorizedError()
        return GrapheneRunNotFoundError(run_id)

    run = record.dagster_run
    graphene_run = GrapheneRun(record)

    if not has_permission_for_run(graphene_info, Permissions.TERMINATE_PIPELINE_EXECUTION, run):
        return GrapheneTerminateRunFailure(
            run=graphene_run,
            message="You do not have permission to terminate this run",
        )

    can_cancel_run = run.status in CANCELABLE_RUN_STATUSES

    valid_status = not run.is_finished and (force_mark_as_canceled or can_cancel_run)

    if not valid_status:
        return GrapheneTerminateRunFailure(
            run=graphene_run,
            message=f"Run {run.run_id} could not be terminated due to having status {run.status.value}.",
        )

    if force_mark_as_canceled:
        try:
            if instance.run_coordinator and can_cancel_run:
                instance.run_coordinator.cancel_run(run_id)
        except:
            instance.report_engine_event(
                "Exception while attempting to force-terminate run. Run will still be marked as"
                " canceled.",
                job_name=run.job_name,
                run_id=run.run_id,
                engine_event_data=EngineEventData(
                    error=serializable_error_info_from_exc_info(sys.exc_info()),
                ),
            )
        return _force_mark_as_canceled(instance, run_id)

    if instance.run_coordinator and can_cancel_run and instance.run_coordinator.cancel_run(run_id):
        return GrapheneTerminateRunSuccess(graphene_run)

    return GrapheneTerminateRunFailure(
        run=graphene_run, message=f"Unable to terminate run {run.run_id}"
    )


def terminate_pipeline_execution_for_runs(
    graphene_info: "ResolveInfo",
    run_ids: Sequence[str],
    terminate_policy: "GrapheneTerminateRunPolicy",
) -> "GrapheneTerminateRunsResult":
    from dagster_graphql.schema.roots.mutation import GrapheneTerminateRunsResult

    check.sequence_param(run_ids, "run_id", of_type=str)

    terminate_run_results = []

    for run_id in run_ids:
        result = terminate_pipeline_execution(
            graphene_info,
            run_id,
            terminate_policy,
        )
        terminate_run_results.append(result)

    return GrapheneTerminateRunsResult(terminateRunResults=terminate_run_results)


def delete_pipeline_run(
    graphene_info: "ResolveInfo", run_id: str
) -> Union["GrapheneDeletePipelineRunSuccess", "GrapheneRunNotFoundError"]:
    from dagster_graphql.schema.errors import GrapheneRunNotFoundError
    from dagster_graphql.schema.roots.mutation import GrapheneDeletePipelineRunSuccess

    instance = graphene_info.context.instance

    run = instance.get_run_by_id(run_id)
    if not run:
        assert_permission(graphene_info, Permissions.DELETE_PIPELINE_RUN)
        return GrapheneRunNotFoundError(run_id)

    assert_permission_for_run(graphene_info, Permissions.DELETE_PIPELINE_RUN, run)

    instance.delete_run(run_id)

    return GrapheneDeletePipelineRunSuccess(run_id)


@deprecated(
    breaking_version="2.0",
    emit_runtime_warning=False,
    additional_warn_text="DAGIT_EVENT_LOAD_CHUNK_SIZE is the only deprecated part.",
)
def get_chunk_size() -> int:
    return int(
        os.getenv(
            "DAGSTER_UI_EVENT_LOAD_CHUNK_SIZE", os.getenv("DAGIT_EVENT_LOAD_CHUNK_SIZE", "1000")
        )
    )


async def gen_events_for_run(
    graphene_info: "ResolveInfo",
    run_id: str,
    after_cursor: str | None = None,
) -> AsyncIterator[
    Union[
        "GraphenePipelineRunLogsSubscriptionFailure",
        "GraphenePipelineRunLogsSubscriptionSuccess",
    ]
]:
    from dagster_graphql.implementation.events import (
        from_event_record,
        get_graphene_events_from_records_connection,
    )
    from dagster_graphql.schema.pipelines.pipeline import GrapheneRun
    from dagster_graphql.schema.pipelines.subscription import (
        GraphenePipelineRunLogsSubscriptionFailure,
        GraphenePipelineRunLogsSubscriptionSuccess,
    )

    check.str_param(run_id, "run_id")
    after_cursor = check.opt_str_param(after_cursor, "after_cursor")
    instance = graphene_info.context.instance
    record = instance.get_run_record_by_id(run_id)

    if not record:
        yield GraphenePipelineRunLogsSubscriptionFailure(
            missingRunId=run_id,
            message=f"Could not load run with id {run_id}",
        )
        return

    run = record.dagster_run

    dont_send_past_records = False
    # special sigil cursor that signals to start watching for updates only after the current point in time
    if after_cursor == "HEAD":
        dont_send_past_records = True
        after_cursor = None

    chunk_size = get_chunk_size()
    # load the existing events in chunks
    has_more = True
    while has_more:
        # run the fetch in a thread since its sync
        connection = await run_in_threadpool(
            instance.get_records_for_run,
            run_id=run_id,
            cursor=after_cursor,
            limit=chunk_size,
        )

        if not dont_send_past_records:
            yield GraphenePipelineRunLogsSubscriptionSuccess(
                run=GrapheneRun(record),
                messages=get_graphene_events_from_records_connection(
                    instance, connection, run.job_name
                ),
                hasMorePastEvents=connection.has_more,
                cursor=connection.cursor,
            )
        has_more = connection.has_more
        after_cursor = connection.cursor

    loop = asyncio.get_event_loop()
    queue: asyncio.Queue[tuple[EventLogEntry, str]] = asyncio.Queue()

    def _enqueue(event: EventLogEntry, cursor: str):
        loop.call_soon_threadsafe(queue.put_nowait, (event, cursor))

    # watch for live events
    instance.watch_event_logs(run_id, after_cursor, _enqueue)
    show_failed_to_materialize = instance.can_read_asset_failure_events()
    try:
        while True:
            event, cursor = await queue.get()
            if (
                show_failed_to_materialize
                or event.dagster_event_type != DagsterEventType.ASSET_FAILED_TO_MATERIALIZE
            ):
                yield GraphenePipelineRunLogsSubscriptionSuccess(
                    run=GrapheneRun(record),
                    messages=[from_event_record(event, run.job_name)],
                    hasMorePastEvents=False,
                    cursor=cursor,
                )
    finally:
        instance.end_watch_event_logs(run_id, _enqueue)


async def gen_captured_log_data(
    graphene_info: "ResolveInfo", log_key: Sequence[str], cursor: str | None = None
) -> AsyncIterator["GrapheneCapturedLogs"]:
    from dagster_graphql.schema.logs.compute_logs import from_captured_log_data

    instance = graphene_info.context.instance

    compute_log_manager = instance.compute_log_manager
    subscription = compute_log_manager.subscribe(log_key, cursor)

    loop = asyncio.get_event_loop()
    queue: asyncio.Queue[CapturedLogData] = asyncio.Queue()

    def _enqueue(new_event):
        loop.call_soon_threadsafe(queue.put_nowait, new_event)

    # subscription object will attempt to fetch when started, so move off main thread
    await run_in_threadpool(subscription, _enqueue)

    is_complete = False
    try:
        while not is_complete:
            update = await queue.get()
            yield from_captured_log_data(update)
            is_complete = subscription.is_complete and queue.empty()
    finally:
        subscription.dispose()


def wipe_assets(
    graphene_info: "ResolveInfo", asset_partition_ranges: Sequence[AssetPartitionWipeRange]
) -> Union[
    "GrapheneAssetWipeSuccess", "GrapheneUnsupportedOperationError", "GrapheneAssetNotFoundError"
]:
    from dagster_graphql.schema.backfill import GrapheneAssetPartitionRange
    from dagster_graphql.schema.errors import (
        GrapheneAssetNotFoundError,
        GrapheneUnsupportedOperationError,
    )
    from dagster_graphql.schema.roots.mutation import GrapheneAssetWipeSuccess

    instance = graphene_info.context.instance
    whole_assets_to_wipe: list[AssetKey] = []
    for apr in asset_partition_ranges:
        if apr.partition_range is None:
            whole_assets_to_wipe.append(apr.asset_key)
        else:
            if apr.asset_key not in graphene_info.context.asset_graph.asset_node_snaps_by_key:
                return GrapheneAssetNotFoundError(asset_key=apr.asset_key)

            node = graphene_info.context.asset_graph.asset_node_snaps_by_key[apr.asset_key]
            partitions_def = check.not_none(node.partitions).get_partitions_definition()
            with partition_loading_context(dynamic_partitions_store=instance):
                partition_keys = partitions_def.get_partition_keys_in_range(apr.partition_range)
            try:
                instance.wipe_asset_partitions(apr.asset_key, partition_keys)

            # NotImplementedError will be thrown if the underlying EventLogStorage does not support
            # partitioned asset wipe.
            except NotImplementedError:
                return GrapheneUnsupportedOperationError(
                    "Partitioned asset wipe is not supported yet."
                )

    instance.wipe_assets(whole_assets_to_wipe)

    result_ranges = [
        GrapheneAssetPartitionRange(asset_key=apr.asset_key, partition_range=apr.partition_range)
        for apr in asset_partition_ranges
    ]
    return GrapheneAssetWipeSuccess(assetPartitionRanges=result_ranges)


def create_asset_event(
    event_type: DagsterEventType,
    asset_key: AssetKey,
    partition_key: str | None,
    description: str | None,
    tags: Mapping[str, str] | None,
) -> AssetMaterialization | AssetObservation:
    if event_type == DagsterEventType.ASSET_MATERIALIZATION:
        return AssetMaterialization(
            asset_key=asset_key, partition=partition_key, description=description, tags=tags
        )
    elif event_type == DagsterEventType.ASSET_OBSERVATION:
        return AssetObservation(
            asset_key=asset_key, partition=partition_key, description=description, tags=tags
        )
    else:
        check.failed(f"Unexpected event type {event_type}")


def report_runless_asset_events(
    graphene_info: "ResolveInfo",
    event_type: DagsterEventType,
    asset_key: AssetKey,
    partition_keys: Sequence[str] | None = None,
    description: str | None = None,
    tags: Mapping[str, str] | None = None,
) -> "GrapheneReportRunlessAssetEventsSuccess":
    from dagster_graphql.schema.roots.mutation import GrapheneReportRunlessAssetEventsSuccess

    instance = graphene_info.context.instance

    if partition_keys is not None:
        for partition_key in partition_keys:
            instance.report_runless_asset_event(
                create_asset_event(event_type, asset_key, partition_key, description, tags)
            )
    else:
        instance.report_runless_asset_event(
            create_asset_event(event_type, asset_key, None, description, tags)
        )

    return GrapheneReportRunlessAssetEventsSuccess(assetKey=asset_key)


def report_asset_check_evaluation(
    graphene_info: "ResolveInfo",
    asset_key: AssetKey,
    check_name: str,
    passed: bool,
    severity: "AssetCheckSeverity",
    metadata: Mapping[str, "RawMetadataValue"] | None = None,
    partition: str | None = None,
) -> "GrapheneReportAssetCheckEvaluationSuccess":
    from dagster_graphql.schema.roots.mutation import GrapheneReportAssetCheckEvaluationSuccess

    instance = graphene_info.context.instance
    evaluation = AssetCheckEvaluation(
        asset_key=asset_key,
        check_name=check_name,
        passed=passed,
        severity=severity,
        metadata=metadata or {},
        partition=partition,
    )
    instance.report_runless_asset_event(evaluation)
    return GrapheneReportAssetCheckEvaluationSuccess(assetKey=asset_key)
